/* ClientController */

#import <Cocoa/Cocoa.h>

#import "WasteApplication.h"

#include "connection.h"

@interface ClientController : NSObject
{
	NSMutableArray	* nickNames;
	NSMutableArray	* channelNames;
	NSMutableArray	* networkConnections;

//	NSLock	* statusLock;
	NSLock	* nickNamesLock;
	NSLock	* channelNamesLock;
	NSLock	* consoleLock;

	NSTimer * scrollTimer;

	IBOutlet id statusText;
	IBOutlet id currentUsersPanel;
	IBOutlet id currentUsersBrowser;
	IBOutlet id channelBrowser;
	IBOutlet id activeSpinner;
	IBOutlet id consoleWindow;
	IBOutlet id consoleTextView;
	IBOutlet id consoleScrollView;

    IBOutlet id joinCreateChatWindow;
    IBOutlet id joinChatRoomNameTextField;

	IBOutlet id browserWindow;
	IBOutlet id browserSearchComboBox;
	IBOutlet id browserSearchButton;
	IBOutlet id browserShowAllUsersButton;
	IBOutlet id browserGoUpDirectoryButton;
	IBOutlet id browserGoBackButton;
	IBOutlet id browserGoForwardButton;
	IBOutlet id browserTableView;
	IBOutlet id browserFilterResultsTextField;

	IBOutlet id transfersWindow;

	IBOutlet id downloadsTableView;
	IBOutlet id downloadsClearWhenCompleteCheckBox;
	IBOutlet id downloadsAutoResumeTimeoutsCheckBox;
	IBOutlet id downloadsMaxDownloadsTextField;
	IBOutlet id downloadsDownloadButton;
	IBOutlet id downloadsStopButton;
	IBOutlet id downloadsClearCompletedButton;
	IBOutlet id downloadsClearAllButton;
	IBOutlet id downloadsClearInactiveButton;

	IBOutlet id downloadQueueTableView;
	IBOutlet id downloadQueueDownloadButton;
	IBOutlet id downloadQueueMoveUpButton;
	IBOutlet id downloadQueueMoveDownButton;
	IBOutlet id downloadQueueStopButton;
	IBOutlet id downloadQueueClearAllButton;
	IBOutlet id downloadQueueItemsInQueueLabel;

	IBOutlet id uploadsTableView;
	IBOutlet id uploadsStopButton;
	IBOutlet id uploadsClearAllButton;
	IBOutlet id uploadsClearInactiveButton;
	IBOutlet id uploadsClearWhenCompleteCheckBox;

    IBOutlet id networkStatusWindow;
    IBOutlet id networkConnectionComboBox;
    IBOutlet id networkMaxConnectionsTextField;
    IBOutlet id networkConnectionsTableView;
	IBOutlet id networkConnectButton;
    IBOutlet id networkConnectSelectedButton;
    IBOutlet id networkDisconnectButton;
    IBOutlet id networkDeleteButton;
}

- (void) gotNick: (NSString *) nick;
- (void) gotChannel: (NSString *) channel;
- (void) startSpinner;
- (void) stopSpinner;
- (void) updateStatus;
- (void) updateServerStatus;
- (void) resetNickNames;
- (void) addStatusItem: (C_Connection *)connectionItem withName: (char *)name withHost: (char *)host withRating: (int)rating withKey: (char *)key;
- (int) getNetworkConnectionCount;
- (void) updateNetworkConnectionWindow;
- (C_Connection *) getNetworkConnectionAtIndex: (int)index;
- (char *)getHostNameAtIndex: (int)index;
- (void) removeNetworkConnectionAtIndex: (int)index;
- (int)getRatingAtIndex: (int)index;
- (BOOL) hasNetworkConnectionAtIndex: (int)index;
- (void) updateNetworkConnectionTable;
- (void) showTransfers;
- (void) showBrowser;
- (void) showNetworkStatus;
- (void) joinChat;
- (void) showConsole;
- (void)scrollToVisible;
- (void)appendToConsole: (NSString *)string;

- (IBAction)joinChat:(id)sender;

- (IBAction)search:(id)sender;
- (IBAction)showAllUsers:(id)sender;
- (IBAction)goUpDirectory:(id)sender;
- (IBAction)goBack:(id)sender;
- (IBAction)goForward:(id)sender;
- (IBAction)filterResults:(id)sender;

- (IBAction)setClearWhenComplete:(id)sender;
- (IBAction)setAutoResulmeTimeouts:(id)sender;
- (IBAction)setMaxDownloads:(id)sender;
- (IBAction)resumeDownload:(id)sender;
- (IBAction)stopDownload:(id)sender;
- (IBAction)clearCompletedDownloads:(id)sender;
- (IBAction)clearAllDownloads:(id)sender;
- (IBAction)clearInactiveDownloads:(id)sender;

- (IBAction)downloadNow:(id)sender;
- (IBAction)moveUp:(id)sender;
- (IBAction)moveDown:(id)sender;
- (IBAction)dequeueDownload:(id)sender;
- (IBAction)dequeueAllDownloads:(id)sender;

- (IBAction)stopUpload:(id)sender;
- (IBAction)clearAllUploads:(id)sender;
- (IBAction)clearInactiveUploads:(id)sender;
- (IBAction)setClearCompletedUploads:(id)sender;

- (IBAction)newNetworkConnection:(id)sender;
- (IBAction)makeNetworkConnection:(id)sender;
- (IBAction)breakNetworkConnection:(id)sender;
- (IBAction)deleteNetworkConnection:(id)sender;
- (IBAction)setMaintainNetworkConnections:(id)sender;

@end
