#import "ClientController.h"
#import "mqueue.h"
#import "m_chat.h"
#import "pubkey.h"
#include "util.h"

#import "ChatController.h"

extern char g_scan_status_buf[128];
extern void msgNode(int numfiles, int numMB);
extern void msgStats();
extern void SaveNetQ(void);
extern void LoadNetQ(void);
extern void update_network_connections();
extern void add_network_connection(char *);
extern void debug_printf(char *, ...);

int g_nodes_added = 0;

@interface NSMutableArray (ObjectForKeyExtension)

- (id)objectWithValue: (id) value forKey: (NSString *)key;

@end

@implementation NSMutableArray (ObjectForKeyExtension)

- (id)objectWithValue: (id) value forKey: (NSString *)key;
{
	id enumerator = [self objectEnumerator];
	id object = nil;
	while (object = [enumerator nextObject]) {
		if ([[(NSDictionary *)object objectForKey: key] isEqual: value]) {
			return object;
		}
	}
	return nil;
}

@end

@implementation ClientController

- (id)init;
{
	self = [super init];

	nickNamesLock = [[NSLock alloc] init];
	channelNamesLock = [[NSLock alloc] init];
//	statusLock = [[NSLock alloc] init];
	consoleLock = [[NSLock alloc] init];

	nickNames = [[NSMutableArray array] retain];
	channelNames = [[NSMutableArray array] retain];
	networkConnections = [[NSMutableArray array] retain];

	scrollTimer = [[NSTimer scheduledTimerWithTimeInterval: 0.5 target: self selector: (SEL)@selector(scrollToVisible) userInfo: nil repeats: YES] retain];

	return self;
}

- (int) getNetworkConnectionCount;
{
	return [networkConnections count];
}

- (void) gotNick: (NSString *) nick;
{
	debug_printf("gotNick: %s\n", [nick cString]);

	NSAssert(nick, @"nick string should not be nil");
	NSAssert(nickNames, @"nickNames array should not be nil");
	NSAssert(currentUsersBrowser, @"currentUsersBrowser control should not be nil");
	NSAssert([currentUsersBrowser delegate] == self, @"currentUsersBrowser delegate should be this object");
	NSAssert(currentUsersPanel, @"currentUsersPanel window should not be nil");

	// add name to nickNames array if not already there
	[nickNamesLock lock];
	if ([nickNames indexOfObject: nick] == NSNotFound) {
		[nickNames addObject: nick];
	}
	[nickNamesLock unlock];
	[currentUsersBrowser reloadColumn: 0];
	//	if ([[[WasteApplication sharedApplication] preferences] objectForKey: @"orderFrontOnUserListUpdate"] == YES) {
	//		[currentUsersPanel makeKeyAndOrderFront: nil];
	//	}
}

- (void) gotChannel: (NSString *) channel;
{
	debug_printf("gotChannel: %s\n", [channel cString]);

	NSAssert(channel, @"channel string should not be nil");
	NSAssert(channelNames, @"channelNames array should not be nil");
	NSAssert(channelBrowser, @"channelBrowser control should not be nil");
	NSAssert([channelBrowser delegate] == self, @"channelBrowser delegate should be this object");
	NSAssert(currentUsersPanel, @"currentUsersPanel window should not be nil");

	// add name to nickNames array if not already there
	if ([channelNames indexOfObject: channel] == NSNotFound) {
		[channelNamesLock lock];
		[channelNames addObject: channel];
		[channelNamesLock unlock];
	}
	[channelBrowser reloadColumn: 0];
 //	[self startSpinner];
}

- (void) startSpinner;
{
	[activeSpinner startAnimation: nil];
}

- (void) stopSpinner;
{
	[activeSpinner stopAnimation: nil];
}

- (void) updateStatus;
{
	if (g_scan_status_buf != NULL && strlen(g_scan_status_buf)) {
		NSString *string = [NSString stringWithCString: g_scan_status_buf];
		[statusText setStringValue: string];
	} else {
//		[statusText setStringValue: @""];
	}
}

- (void) updateServerStatus;
{
//	[statusLock lock];
	NSColor * color = [statusText textColor];
	id calibratedColor = nil;
	NSColor * newColor = nil;
	float white = 0.0, alpha = 0.0;
	static float increment = 0.01;
	static int wait = 0;
	int wait_time = 80;	// how long the text "waits" at the black color
	calibratedColor = [color colorUsingColorSpaceName: NSDeviceWhiteColorSpace];
	[calibratedColor getWhite: &white alpha: &alpha];
	if (wait > 0) {
		wait--;
	} else {
		white += increment;
	}
	if (white >= 1.0) {
		white = 1.0;
		increment = -increment;
		white += increment;
		if (wait <= 0) wait = 1;
	}
	if (white <= 0.0) {
		white = 0.0;
		increment = -increment;
		white += increment;
		if (wait <= 0) wait = wait_time;
	}
	if (wait == wait_time) {
		[currentUsersBrowser reloadColumn: 0];
	}
	newColor = [NSColor colorWithCalibratedWhite: white alpha: alpha];
	[statusText setTextColor: newColor];
	NS_DURING
		[self updateNetworkConnectionWindow];
	NS_HANDLER
		NSLog(@"AN EXCEPTION WAS RAISED - %@", localException);
	NS_ENDHANDLER
//	[statusLock unlock];
}

- (void) updateNetworkConnectionWindow;
{
	update_network_connections();
}

- (C_Connection *) getNetworkConnectionAtIndex: (int)index;
{
	if ((unsigned)index < [networkConnections count]) {
		return (C_Connection *)[[[networkConnections objectAtIndex: index] objectForKey: @"connection"] intValue];
	}
	return NULL;
}

- (void) removeNetworkConnectionAtIndex: (int)index;
{
	return;
	if ((unsigned)index < [networkConnections count]) [networkConnections removeObjectAtIndex: index];
}

- (char *)getHostNameAtIndex: (int)index;
{
	if ((unsigned)index < [networkConnections count]) {
		return (char *)[[[networkConnections objectAtIndex: index] objectForKey: @"host"] cString];
	}
	return NULL;
}

- (int)getRatingAtIndex: (int)index;
{
	if ((unsigned)index < [networkConnections count]) {
		return (int)[[[networkConnections objectAtIndex: index] objectForKey: @"rating"] intValue];
	}
	return -1;
}

- (BOOL) hasNetworkConnectionAtIndex: (int)index;
{
	return (BOOL)(((unsigned)index < [networkConnections count]) && [networkConnections objectAtIndex: index] != nil);
}

- (void) updateNetworkConnectionTable;
{
	[networkConnectionsTableView reloadData];
}

- (void) addStatusItem: (C_Connection *)connectionItem withName: (char *)name withHost: (char *)host withRating: (int)rating withKey: (char *)key;
{
//	debug_printf("addStatusItem: %d withItem: %s withHost: %s withRating: %d\n", (int)connectionItem, name, host, rating);
	NSString * nameString = [NSString stringWithCString: name];
	NSString * hostName = [NSString stringWithCString: host];
	NSMutableDictionary * dict = [networkConnections objectWithValue: hostName forKey: @"host"];
	if (dict) {
		[dict setObject: nameString forKey: @"status"];
		[dict setObject: [NSNumber numberWithInt: rating] forKey: @"rating"];
		[dict setObject: [NSString stringWithCString: key] forKey: @"user"];
		NSLog(@"Found connection for host : %s", host);
	} else {
		dict = [NSMutableDictionary dictionary];
		[dict setObject: hostName forKey: @"host"];
		[dict setObject: nameString forKey: @"status"];
		[dict setObject: [NSNumber numberWithInt: rating] forKey: @"rating"];
		[dict setObject: [NSString stringWithCString: key] forKey: @"user"];
		[dict setObject: [NSNumber numberWithInt: (int)connectionItem] forKey: @"connection"];
		[networkConnections addObject: dict];
		NSLog(@"Couldn't find connection for host : %s", host);
	}
	[networkConnectionsTableView reloadData];
}

- (void)resetNickNames;
{
	[nickNamesLock lock];
	[nickNames removeAllObjects];
	[nickNamesLock unlock];
}


- (void)awakeFromNib;
{
	[[consoleTextView textContainer] setWidthTracksTextView: YES];

	// Set activity spinner style
	[activeSpinner setIndeterminate: YES];
	[activeSpinner setDisplayedWhenStopped: NO];
	[activeSpinner setStyle: NSProgressIndicatorSpinningStyle];
	[activeSpinner setUsesThreadedAnimation: YES];

	// Clear status bar text
//	[statusText setStringValue: @""];

	// Setup delegate and datasource for Network Connections table view
	[networkConnectionsTableView setDelegate: self];
	[networkConnectionsTableView setDataSource: self];
	[networkConnectionsTableView setAutosaveName: @"netcontableview"];
	[networkConnectionsTableView setAutoresizesAllColumnsToFit: NO];
	[networkConnectionsTableView setAutosaveTableColumns: YES];

	// Make sure the NSBrowser delegate is set to this object
	[currentUsersBrowser setDelegate: self];

	[currentUsersPanel makeKeyAndOrderFront: nil];
}

- (void) showConsole;
{
	[consoleWindow makeKeyAndOrderFront: nil];
	if (![scrollTimer isValid]) {
		[scrollTimer fire];
	}
}

- (void)scrollToVisible;
{
	int l = [[consoleTextView string] length];
	if (l) {
		[consoleTextView scrollRangeToVisible:NSMakeRange(l, 0)];
	}
//	if ([consoleTextView string]) {
//		[consoleTextView scrollRangeToVisible: NSMakeRange([[consoleTextView string] length]-1, 1)];
//		[consoleTextView setNeedsDisplay: YES];
//		[consoleScrollView setNeedsDisplay: YES];
//		[consoleScrollView reflectScrolledClipView: [consoleScrollView contentView]];
//	}
}

- (void)appendToConsole: (NSString *)string;
{
	[consoleLock lock];
	[[consoleTextView textStorage] setFont: [NSFont fontWithName: @"Courier" size: 12.0]];
    [[consoleTextView textStorage] appendAttributedString: [[[NSAttributedString alloc] initWithString: string] autorelease]];
//    [consoleTextView setString: [NSString stringWithFormat: @"%@%@", [consoleTextView string], string]];
	[consoleLock unlock];
//    [self performSelector: (SEL)@selector(scrollToVisible) withObject: nil afterDelay: 0.1];
}

/////////////////////////////////////////////////////////////////////////////////////
// Interface Builder action methods for all of the UI stuff
/////////////////////////////////////////////////////////////////////////////////////


- (IBAction)joinChat:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - joinChat: %@", [self class], sender);
#endif
	[[[WasteApplication sharedApplication] chatController] createNewChatRoomWithChannel: [sender stringValue]];
	if ([[sender window] isEqual: joinCreateChatWindow]) {
		[[sender window] performClose: nil];
	}
}

- (void)joinChat;
{
#ifdef _DEBUG_
	NSLog(@"%@ - joinChat", [self class]);
#endif
	[joinCreateChatWindow makeKeyAndOrderFront: nil];
}


- (IBAction)search:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - search: %@", [self class], sender);
#endif
}

- (void)showTransfers;
{
	[transfersWindow makeKeyAndOrderFront: nil];
}

- (void)showBrowser;
{
	[browserWindow makeKeyAndOrderFront: nil];
}

- (void)showNetworkStatus;
{
	[networkStatusWindow makeKeyAndOrderFront: nil];
}

- (IBAction)showAllUsers:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - showAllUsers: %@", [self class], sender);
#endif
}

- (IBAction)goUpDirectory:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - goUpDirectory: %@", [self class], sender);
#endif
}

- (IBAction)goBack:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - goBack: %@", [self class], sender);
#endif
}

- (IBAction)goForward:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - goForward: %@", [self class], sender);
#endif
}

- (IBAction)filterResults:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - filterResults: %@", [self class], sender);
#endif
}


- (IBAction)setClearWhenComplete:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - setClearWhenComplete: %@", [self class], sender);
#endif
}

- (IBAction)setAutoResulmeTimeouts:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - setAutoResulmeTimeouts: %@", [self class], sender);
#endif
}

- (IBAction)setMaxDownloads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - setMaxDownloads: %@", [self class], sender);
#endif
}

- (IBAction)resumeDownload:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - resumeDownload: %@", [self class], sender);
#endif
}

- (IBAction)stopDownload:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - stopDownload: %@", [self class], sender);
#endif
}

- (IBAction)clearCompletedDownloads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - clearCompletedDownloads: %@", [self class], sender);
#endif
}

- (IBAction)clearAllDownloads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - clearAllDownloads: %@", [self class], sender);
#endif
}

- (IBAction)clearInactiveDownloads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - clearInactiveDownloads: %@", [self class], sender);
#endif
}

- (IBAction)downloadNow:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - downloadNow: %@", [self class], sender);
#endif
}

- (IBAction)moveUp:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - moveUp: %@", [self class], sender);
#endif
}

- (IBAction)moveDown:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - moveDown: %@", [self class], sender);
#endif
}

- (IBAction)dequeueDownload:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - dequeueDownload: %@", [self class], sender);
#endif
}

- (IBAction)dequeueAllDownloads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - dequeueAllDownloads: %@", [self class], sender);
#endif
}

- (IBAction)stopUpload:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - stopUpload: %@", [self class], sender);
#endif
}

- (IBAction)clearAllUploads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - clearAllUploads: %@", [self class], sender);
#endif
}

- (IBAction)clearInactiveUploads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - clearInactiveUploads: %@", [self class], sender);
#endif
}

- (IBAction)setClearCompletedUploads:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - setClearCompletedUploads: %@", [self class], sender);
#endif
}


- (IBAction)newNetworkConnection:(id)sender;
{
	debug_printf("newNetworkConnection: %s\n", [[sender stringValue] cString]);
	if ([sender stringValue]) {
		add_network_connection((char *)[[sender stringValue] cString]);
		[networkConnectionComboBox addItemWithObjectValue: [sender stringValue]];
	}
}

- (IBAction)makeNetworkConnection:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - makeNetworkConnection: %@", [self class], sender);
#endif
}

- (IBAction)breakNetworkConnection:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - breakNetworkConnection: %@", [self class], sender);
#endif
}

- (IBAction)deleteNetworkConnection:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - deleteNetworkConnection: %@", [self class], sender);
#endif
}

- (IBAction)setMaintainNetworkConnections:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - setMaintainNetworkConnections: %@", [self class], sender);
#endif
}

//////////////////////////////////////////////////////////////////////////////
// NSBrowser delegate methods
//////////////////////////////////////////////////////////////////////////////

- (int)browser: (NSBrowser *)browser numberOfRowsInColumn: (int)column;
{
	int ret = 0;

#ifdef _DEBUG_
	NSLog(@"%@ - browser: %@ numberOfRowsInColumn: %d", [self class], browser, column);
#endif

	if ([browser tag] == 0) {
		NSAssert(nickNames, @"nickNames array should not be nil");
		ret = [nickNames count];
	}
	if ([browser tag] == 1) {
		NSAssert(channelNames, @"channelNames array should not be nil");
		ret = [channelNames count];
	}

	NSLog(@"number of columns = %d", ret);

	return ret;
}

- (void)browser: (NSBrowser *)browser willDisplayCell: (id)cell atRow: (int)row column: (int)column;
{

#ifdef _DEBUG_
	NSLog(@"%@ - browser: %@ willDisplayCell: %@ atRow: %d column: %d", [self class], browser, cell, row, column);
#endif

	[cell setLeaf: YES];

	if ([browser tag] == 0) {
		NSAssert(nickNames != nil, @"nickNames array should not be nil");
		if ((unsigned)row < [nickNames count] - 1) {
			return;
		}
		NSAssert((unsigned)row < [nickNames count], ([NSString stringWithFormat: @"row (%d) is out of bounds of nickNames array", row]));
		NSString * nick = [nickNames objectAtIndex: row];
		NSAssert(nick != nil, @"nick string should not be nil");
		[cell setStringValue: nick];
		if ([cell image] == nil) {
			[cell setImage: [NSImage imageNamed: @"user.ico"]];
		}
	}

	if ([browser tag] == 1) {
		NSAssert(channelNames != nil, @"channelNames array should not be nil");
		if ((unsigned)row < [channelNames count] - 1) {
			return;
		}
		NSAssert((unsigned)row < [channelNames count], ([NSString stringWithFormat: @"row (%d) is out of bounds of channelNames array", row]));
		NSString * channel = [channelNames objectAtIndex: row];
		NSAssert(channel != nil, @"channel string should not be nil");
		[cell setStringValue: channel];
		if ([cell image] == nil) {
			[cell setImage: [NSImage imageNamed: @"allusers.ico"]];
		}
	}
}

- (void)dealloc;
{
#ifdef _DEBUG_
	NSLog(@"%@ - dealloc", [self class]);
#endif
	[nickNamesLock unlock];
	[nickNamesLock release];
	[channelNamesLock unlock];
	[channelNamesLock release];
	[consoleLock unlock];
	[consoleLock release];
	[nickNames release];
	[channelNames release];
	[networkConnections release];
	[scrollTimer invalidate];
	[scrollTimer release];
	[super dealloc];
}

//////////////////////////////////////////////////////////////////////////////
// NSTableView datasource methods
//////////////////////////////////////////////////////////////////////////////

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
#ifdef _DEBUG_
//	NSLog(@"%@ - numberOfRowsInTableView: %@", [self class], aTableView);
#endif
	if ([aTableView tag] == 0) {
		NSAssert(networkConnections, @"networkConnections array should not be nil");
		return [networkConnections count];
	}
	if ([aTableView tag] == 1) {
//		NSAssert(pendingKeys, @"pendingKeys array should not be nil");
//		return g_pklist_pending.GetSize();
	}
	if ([aTableView tag] == 2) {
//		NSAssert(ipAddresses, @"ipAddresses array should not be nil");
//		return [ipAddresses count];
	}
	return 0;
}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)row
{
#ifdef _DEBUG_
//	NSLog(@"%@ - tableView: %@ objectValueForTableColumn: %@ row: %d", [self class], aTableView, aTableColumn, row);
#endif
	if ([aTableView isEqual: networkConnectionsTableView]) {
		NSAssert(networkConnections, @"networkConnections array should not be nil");
		if ((unsigned)row < [networkConnections count] - 1) {
			return nil;
		}
		NSAssert((unsigned)row < [networkConnections count], ([NSString stringWithFormat: @"row (%d) is out of bounds of networkConnections array", row]));
		NSDictionary * netCon = [networkConnections objectAtIndex: row];
		if ([[aTableColumn identifier] intValue] == 0) {
			return [netCon objectForKey: @"status"];
		}
		if ([[aTableColumn identifier] intValue] == 1) {
			return [netCon objectForKey: @"host"];
		}
		if ([[aTableColumn identifier] intValue] == 2) {
			return [netCon objectForKey: @"rating"];
		}
		if ([[aTableColumn identifier] intValue] == 3) {
			return [netCon objectForKey: @"user"];
		}
	}
	if ([aTableView isEqual: browserTableView]) {
		if ([[aTableColumn identifier] intValue] == 1) {
   //		NSAssert(pendingKeys != nil, @"pendingKeys array should not be nil");
   //		NSString *key = [pendingKeys objectAtIndex: row];
   //		NSAssert(key != nil, @"key string should not be nil");
   //		return key;
   //		[cell setStringValue: key];
   //		if ([cell image] == nil) {
   //			[cell setImage: [NSImage imageNamed: @"allusers.ico"]];
   //		}
		}
	}

	if ([aTableView isEqual: downloadsTableView]) {
		if ([[aTableColumn identifier] intValue] == 2) {
   //		NSAssert(ipAddresses != nil, @"ipAddresses array should not be nil");
   //		NSString *ip = [ipAddresses objectAtIndex: row];
   //		NSAssert(ip != nil, @"channel string should not be nil");
   //		return ip;
   //		[cell setStringValue: ip];
   //		if ([cell image] == nil) {
   //			[cell setImage: [NSImage imageNamed: @"allusers.ico"]];
   //		}
		}
	}
	return @"";
}

//////////////////////////////////////////////////////////////////////////////
// NSTabView delegate methods
//////////////////////////////////////////////////////////////////////////////

- (BOOL)tabView:(NSTabView *)tabView shouldSelectTabViewItem:(NSTabViewItem *)tabViewItem;
{
	return YES;
	if (![(NSString *)[tabViewItem identifier] length]) {
		return NO;
	}
	switch ([[tabViewItem identifier] intValue]) {
//		case 0: [publicKeyTableView reloadData]; break;
//		case 1: [pendingKeyTableView reloadData]; break;
//		case 2: [ipAddressTableView reloadData]; break;
	}
	return YES;
}
































////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
WASTE - main.cpp (Windows main entry point and a lot of code :)
 Copyright (C) 2003 Nullsoft, Inc.

 WASTE is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 WASTE  is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with WASTE; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
---- some things that would be good to do eventually:

	permissions per key:
    C=chat
    S=search/browse out
    s=search/browsable
    X=file transfer
    K=key distribution
    U=uploads
    P=pings/ips/etc
    ?=unknown (all others)
	have the connection keep track of the flags and key hash, then have the
	prefs update all connections' flags.

	improve ip list sorting (use a sorted listview?)

	plugin interface
	ipv6 support
	Ephemeral Diffie-Hellman for session key exchange?

---- bits about how the negotiation works:

	RNG: based on RSAREF's, but extended (uses 32 byte state, with 16 bytes counter, and 16 bytes of system entropy constantly mixed in --
	this takes from when messages arrive, mouse movement, timing information on when connections come up, etc).

	On connection:

	1)
	A sends B 16 random bytes (bfhpkA)
	A sends B blowFish(bfhpkA,20 byte SHA-1 of public key + 4 pad bytes)
	B sends A 16 random bytes (bfhpkB)
	B sends A blowFish(bfhpkB,20 byte SHA-1 of public key + 4 pad bytes)

	2)
	if A knows B's public key, and B knows A public key, continue.
	A sends B: RSA(pubkey_B,sKey1 + bfhpkB), sKey1 is 64 random bytes. (+ = concat)
	B sends A: RSA(pubKey_A,sKey2 + bfhpkA), sKey2 is 64 random bytes. (+ = concat)

	Check to make sure bfhpkA and bfhpkB are correct, and to make sure sKey1^sKey2 is nonzero,
	with the last 8 bytes of sKey1 differing from the last 8 bytes of sKey2.

	3)
	each side initializes blowfish using the first 56 bytes sKey1^sKey2 as key.
	A uses the last 8 bytes of sKey2 as recv CBC IV.
	B uses the last 8 bytes of sKey1 as recv CBC IV.

	4) 
	each side sends a 16 byte signature, blowfished.  
	each side verifies 16 byte signature.
	message communication begins.
	further messages are verified with a MD5 to detect tampering.

*/

extern void FormatSizeStr64(char *out, unsigned int low, unsigned int high);

#include "main.h"
#include "md5.h"

#include "m_upload.h"
#include "m_chat.h"
#include "m_search.h"
#include "m_ping.h"
#include "m_file.h"
#include "m_lcaps.h"
#include "m_keydist.h"
#include "netq.h";
#include "netkern.h"
#include "xfers.h"
#include "xferwnd.h"

#define VERSION "v1.0b"

int g_exit=0;
int g_has_started=0;
int g_scanworking=0;

char *g_nameverstr=APP_NAME " " VERSION;
char *g_def_extlist="ppt;doc;xls;txt;zip;";

C_FileDB *g_database, *g_newdatabase;
C_AsyncDNS *g_dns;
C_Listen *g_listen;
C_MessageQueueList *g_mql;
C_Config *g_config;

//HMENU g_context_menus;
//HWND g_mainwnd;
//HICON g_hSmallIcon;
//HINSTANCE g_hInst;

int g_hidewnd_state;

unsigned char g_networkhash[SHA_OUTSIZE];
int g_use_networkhash;

char g_config_prefix[1024];
char g_config_mainini[1024];

char g_profile_name[128];
int g_extrainf;
int g_keepup;
int g_conspeed,g_route_traffic;
int g_do_log;
int g_max_simul_dl;
unsigned int g_max_simul_dl_host;
unsigned int g_forceip, g_forceip_addr;
int g_use_accesslist;
int g_appendprofiletitles;
int g_do_autorefresh;
int g_accept_downloads;
int g_port;
int g_chat_timestamp;
int g_keydist_flags;

R_RSA_PRIVATE_KEY g_key;
unsigned char g_pubkeyhash[SHA_OUTSIZE];

char g_regnick[32];

char g_filedlg_ulpath[256];

int g_throttle_flag, g_throttle_send, g_throttle_recv;

int g_search_showfull,g_search_showfullbytes;

int g_scanloadhack;
char g_scan_status_buf[128];
time_t g_next_refreshtime;

time_t g_last_pingtime,g_last_bcastkeytime,g_last_bpstime;
T_GUID g_last_scanid;
int g_last_scanid_used;

T_GUID g_last_pingid;
int g_last_pingid_used;
T_GUID g_client_id;
char g_client_id_str[33];

static C_ItemList<C_UploadRequest> uploadPrompts;
static C_ItemList<C_KeydistRequest> keydistPrompts;

void main_BroadcastPublicKey(T_Message *src)
{
	C_KeydistRequest rep;
	debug_printf("main_BroadcastPublicKey() nick = %s\n", g_regnick);
	rep.set_nick(g_regnick);
	rep.set_flags((g_port && g_listen && !g_listen->is_error())?M_KEYDIST_FLAG_LISTEN:0);
	R_RSA_PUBLIC_KEY k;
	k.bits=g_key.bits;
	memcpy(k.exponent,g_key.publicExponent,MAX_RSA_MODULUS_LEN);
	memcpy(k.modulus,g_key.modulus,MAX_RSA_MODULUS_LEN);
	rep.set_key(&k);

	T_Message msg={{0},0};
	msg.data=rep.Make();
	if (msg.data)
	{
		if (src)
		{
			msg.message_guid=src->message_guid;
			msg.message_type=MESSAGE_KEYDIST_REPLY;
		}
		else
		{
			msg.message_type=MESSAGE_KEYDIST;
		}
		msg.message_length=msg.data->GetLength();
		g_mql->send(&msg);
	}
}

static void main_handleKeyDist(C_KeydistRequest *kdr, int pending)
{
	debug_printf("main_handleKeyDist()\n");
	if (!kdr->get_key()->bits) return;
	if (findPublicKeyFromKey(kdr->get_key())) return;

	PKitem *p=(PKitem *)malloc(sizeof(PKitem));

	SHAify m;
	safe_strncpy(p->name,kdr->get_nick(),sizeof(p->name));
	p->pk = *kdr->get_key();
	m.add((unsigned char *)p->pk.modulus,MAX_RSA_MODULUS_LEN);
	m.add((unsigned char *)p->pk.exponent,MAX_RSA_MODULUS_LEN);
	m.final(p->hash);

	if (pending) 
		g_pklist_pending.Add(p);
	else 
		g_pklist.Add(p);

	savePKList();
}

static void main_handleUpload(char *guidstr, char *fnstr, C_UploadRequest *t)
{
	int willq = Xfer_WillQ(fnstr,guidstr);
	debug_printf("main_handleUpload() guidstr = %s, fnstr = %s, willq = %d\n", guidstr, fnstr, willq);
//	int p=g_lvrecvq.InsertItem(g_lvrecvq.GetCount(),fnstr,0);
	char sizebuf[64];
	strcpy(sizebuf,"?");
	int fs_l,fs_h;
	t->get_fsize(&fs_l,&fs_h);
	if (fs_l != -1 || fs_h != -1)
	{
		FormatSizeStr64(sizebuf,fs_l,fs_h);
	}

//	g_lvrecvq.SetItemText(p,1,sizebuf);
//	g_lvrecvq.SetItemText(p,2,guidstr);
	debug_printf("main_handleUpload() - sizebuf= %s\n", sizebuf);
	debug_printf("main_handleUpload() - guidstr = %s\n", guidstr);
	g_files_in_download_queue++;
	if (g_config->ReadInt("aorecv",1))
	{
//		HWND h=GetForegroundWindow();
//		SendMessage(g_mainwnd,WM_COMMAND,ID_VIEW_TRANSFERS,0);

//		if (g_config->ReadInt("aorecv_btf",0)) SetForegroundWindow(g_xferwnd);
//		else SetForegroundWindow(h);

//		XferDlg_SetSel(willq);
	}
  RecvQ_UpdateStatusText();
}

void main_MsgCallback(T_Message *message, C_MessageQueueList *mql, C_Connection *cn)
{
	id data = [NSData dataWithBytes: message length: message->message_length];
	id string = [[NSMutableString alloc] initWithData: data encoding: [NSString defaultCStringEncoding]];
	debug_printf("main_MsgCallback() message_type = %d, message = %s\n", message->message_type, [string cString]);
	switch(message->message_type)
	{
		case MESSAGE_LOCAL_SATURATE:
			debug_printf("got a %d byte saturation message\n",40+message->message_length);
			break;
		case MESSAGE_LOCAL_CAPS:
		{
			C_MessageLocalCaps mlc(message->data);
			int x;
			const char *host=cn?cn->get_host():"?";
			for (x = 0; x < mlc.get_numcaps(); x ++)
			{
				int n,v;
				mlc.get_cap(x,&n,&v);
				switch (n)
				{
					case MLC_SATURATION:
						debug_printf("%s requests saturation = %d\n",host,v);
						if (cn) cn->set_saturatemode(v);
					//	if ((get_saturatemode()&1) && (g_throttle_flag&32)) do outbound saturation
							break;
					case MLC_BANDWIDTH:
						debug_printf("%s requests max sendbuf = %d\n",host,v);
						if (cn) cn->set_max_sendsize(v);
							break;
				}
			}
		}
			break;
		case MESSAGE_KEYDIST_REPLY:
		case MESSAGE_KEYDIST:

		{
			C_KeydistRequest *r = new C_KeydistRequest(message->data);
			log_printf(1,"recv KEYDIST%s %dB nick='%s' listen=%d\n",
			  (message->message_type==MESSAGE_KEYDIST)?"":"_REPLY",
			  message->message_length, r->get_nick(),
			  r->get_flags()&M_KEYDIST_FLAG_LISTEN);
			if ((r->get_flags() & M_KEYDIST_FLAG_LISTEN) || (g_port && g_listen && !g_listen->is_error()) && r->get_key()->bits)
			{
				if (g_keydist_flags&1) // add without prompt
				{
					main_handleKeyDist(r,0);
					delete r;
				}
				else if (g_keydist_flags&2) // add with prompt
				{
					keydistPrompts.Add(r);
				}
				else // add to pending
				{
					main_handleKeyDist(r,1);
					delete r;
				}

				if (message->message_type != MESSAGE_KEYDIST_REPLY)
				{
					main_BroadcastPublicKey(message);
				}
			}
			else delete r;
		}
			break;
		case MESSAGE_PING:
		{
			MYSRANDUPDATE((unsigned char *)&message->message_guid,16);
			C_MessagePing rep(message->data);
			char ipbuf[32];
			ipstr(ipbuf, rep.m_ip);
			log_printf(1,"recv PING %dB nick='%s' %s:%d\n",
			  message->message_length, rep.m_nick, ipbuf, rep.m_port);
			if (rep.m_nick[0] && rep.m_nick[0] != '#' && rep.m_nick[0] != '&' && rep.m_nick[0] != '.' && strlen(rep.m_nick)<24)
				main_onGotNick(rep.m_nick,0);
		}
			break;
		case MESSAGE_SEARCH_USERLIST:
			log_printf(1,"recv SEARCH_USERLIST %dB\n", message->message_length);
			if (g_regnick[0] && g_regnick[0] != '.' && strlen(g_regnick)<24)
			{
				C_MessageSearchReply repl;
				repl.set_conspeed(g_conspeed);
				repl.set_guid(&g_client_id);
				repl.add_item(-1,g_regnick,"Node",g_database->GetNumFiles(),g_database->GetNumMB(),g_database->GetLatestTime());
				T_Message msg={{0},0};
				msg.message_guid=message->message_guid;
				msg.data=repl.Make();
				if (msg.data)
				{
					msg.message_type=MESSAGE_SEARCH_REPLY;
					msg.message_length=msg.data->GetLength();
					mql->send(&msg);
				}
			}
			break;
		case MESSAGE_SEARCH:
			if (g_accept_downloads&1)
			{
				C_MessageSearchRequest req(message->data);
				char *ss=req.get_searchstring();
				log_printf(1,"recv SEARCH %dB '%s'\n", message->message_length, ss);
				if (g_database->GetNumFiles()>0) {
					if ((ss[0] == '/' && (g_accept_downloads&4)) ||
		 (ss[0] && ss[0] != '/' && (g_accept_downloads&2))) {
						C_MessageSearchReply repl;
						repl.set_conspeed(g_conspeed);
						repl.set_guid(&g_client_id);
						g_database->Search(ss,&repl,mql,message,main_MsgCallback);
					}
				}
			}
			break;
		case MESSAGE_SEARCH_REPLY:
		{
			log_printf(1,"recv SEARCH_REPLY %dB\n", message->message_length);
			if (g_last_scanid_used && !memcmp(&g_last_scanid,&message->message_guid,16))
			{
				C_MessageSearchReply repl(message->data);
				if (repl.get_numitems()==1)
				{
					char name[1024],metadata[256];
					int numfiles,numMB,latestTime;
					if (!repl.get_item(0,NULL,name,metadata,&numfiles,&numMB,&latestTime))
						main_onGotNick(name,0);
					log_printf(1,"  nick='%s' '%s' #files=%d MB=%d\n",
				name, metadata, numfiles, numMB);
					msgNode(numfiles, numMB);
				}
			}
			//      else Search_AddReply(message);
		}
			break;
		case MESSAGE_FILE_REQUEST:
		{
			C_FileSendRequest *r = new C_FileSendRequest(message->data);
			size_t i;
			char buf[64];
			for (i=0;i<sizeof(g_client_id);i++)
				sprintf(buf+2*i, "%02X", r->get_guid()->idc[i]);
			buf[2*i] = '\0';
			log_printf(2,"recv FILE_REQUEST %dB %s\n",
			  message->message_length, buf);
			if (!memcmp(r->get_guid(),&g_client_id,sizeof(g_client_id)))
			{
				int n=g_sends.GetSize();
				int x;
				for (x = 0; x < n; x ++)
				{
					if (!memcmp(r->get_prev_guid(),g_sends.Get(x)->get_guid(),16))
					{
						if (r->is_abort()==2)
						{
							int a=g_sends.Get(x)->GetIdx()-UPLOAD_BASE_IDX;
							if (a >= 0 && a < g_uploads.GetSize())
							{
								char *p=g_uploads.Get(a);
								if (p)
								{
//									int lvidx;
//									while ((lvidx=g_lvsend.FindItemByParam((int)p)) >= 0)
//									{
//										g_lvsend.DeleteItem(lvidx);
//									}
									free(p);
									g_uploads.Set(a,NULL);
								}
							}
						}

						g_sends.Get(x)->set_guid(&message->message_guid);
						g_sends.Get(x)->onGotMsg(r);
						break;
					}
				}
				if (x == n && !r->is_abort()) // new file request
				{
					if (g_config->ReadInt("limit_uls",1) && n < g_config->ReadInt("ul_limit",160))
					{
						char fn[2048];
						fn[0]=0;
						int idx=r->get_idx();
						if (idx >= 0)
						{
							if (idx < UPLOAD_BASE_IDX)
							{
								if (!(g_accept_downloads&1)) fn[0]=0;
								else g_database->GetFile(idx,fn,NULL,NULL,NULL);
							}
							else
							{
								idx-=UPLOAD_BASE_IDX;
								if (idx<g_uploads.GetSize() && g_uploads.Get(idx)) 
								{
									safe_strncpy(fn,g_uploads.Get(idx),sizeof(fn));
//									int lvidx;
//									if ((lvidx=g_lvsend.FindItemByParam((int)g_uploads.Get(idx))) >= 0)
//									{
//										g_lvsend.DeleteItem(lvidx);
//									}
								}
							}
						}
						if (fn[0])
						{
							XferSend *a = new XferSend(mql,&message->message_guid,r,fn);

							// remove any timed out files of a->GetName() from r->get_nick()
//							n=g_lvsend.GetCount();
							for (x = 0; x < n; x ++)
							{
//								if (g_lvsend.GetParam(x)) continue;
								char buf[1024];
//								g_lvsend.GetText(x,3,buf,sizeof(buf));
								if (strncmp(buf,"Timed out",9)) continue;
//								g_lvsend.GetText(x,0,buf,sizeof(buf));
								if (strcmp(buf,a->GetName())) continue;
//								g_lvsend.GetText(x,1,buf,sizeof(buf));
								if (strcmp(buf,r->get_nick())) continue;
//								g_lvsend.DeleteItem(x);
								debug_printf("main_MsgCallback() - r->get_nick() = %s\n", r->get_nick());
								debug_printf("main_MsgCallback() - buf = %s\n", buf);
								debug_printf("main_MsgCallback() - err = Timed out\n");
								x--;
								n--;
							}

							char *err=a->GetError();
							if (err)
							{
								if (!g_config->ReadInt("send_autoclear",0))
								{
//									g_lvsend.InsertItem(0,a->GetName(),0);
									char buf[32];
									int fs_l = 0, fs_h = 0;
//									a->GetSize((unsigned int *)&fs_l,(unsigned int *)&fs_h);
									FormatSizeStr64(buf,fs_l,fs_h);
//									g_lvsend.SetItemText(0,1,r->get_nick());
//									g_lvsend.SetItemText(0,2,buf);
//									g_lvsend.SetItemText(0,3,err);
									debug_printf("main_MsgCallback() - r->get_nick() = %s\n", r->get_nick());
									debug_printf("main_MsgCallback() - buf = %s\n", buf);
									debug_printf("main_MsgCallback() - err = %s\n", err);
								}
								delete a;
							}
							else
							{
								g_sends.Add(a);
//								g_lvsend.InsertItem(0,a->GetName(),(int)a);
								char buf[32];
								int fs_l,fs_h;
								a->GetSize((unsigned int *)&fs_l,(unsigned int *)&fs_h);
								FormatSizeStr64(buf,fs_l,fs_h);
//								g_lvsend.SetItemText(0,1,r->get_nick());
//								g_lvsend.SetItemText(0,2,buf);
//								g_lvsend.SetItemText(0,3,"Sending");
								debug_printf("main_MsgCallback() - r->get_nick() = %s\n", r->get_nick());
								debug_printf("main_MsgCallback() - buf = %s\n", buf);
								debug_printf("main_MsgCallback() - err = Sending\n");
							}
//							// "adjust UI" PostMessage(g_xferwnd,WM_USER_TITLEUPDATE,0,0);
						}
						else
						{
							T_Message msg={{0},0};
							C_FileSendReply reply;
							reply.set_error(1);
							msg.data=reply.Make();
							if (msg.data)
							{
								msg.message_type=MESSAGE_FILE_REQUEST_REPLY;
								msg.message_length=msg.data->GetLength();
								msg.message_guid=message->message_guid;
								g_mql->send(&msg);
							}
						}
					}
				}
			}
			delete r;
		}
			break;
		case MESSAGE_FILE_REQUEST_REPLY:
		{
			log_printf(1,"recv FILE_REQUEST_REPLY %dB\n", message->message_length);
			int n=g_recvs.GetSize();
			int x;
			for (x = 0; x < n; x ++)
			{
				if (!memcmp(g_recvs.Get(x)->get_guid(),&message->message_guid,16))
				{
					g_recvs.Get(x)->onGotMsg(new C_FileSendReply(message->data));
					break;
				}
			}
		}
			break;
		case MESSAGE_CHAT_REPLY:
			log_printf(1,"recv CHAT_REPLY %dB\n", message->message_length);
			[[[WasteApplication sharedApplication] chatController] handleReplyMessage: message];
			break;
		case MESSAGE_CHAT:
			if ([[[WasteApplication sharedApplication] chatController] handleMessage: message] && g_regnick[0])
			{ // send reply
				C_MessageChatReply rep;	  
				rep.setnick(g_regnick);

				T_Message msg={{0},0};
				msg.message_guid=message->message_guid;
				msg.data=rep.Make();
				if (msg.data)
				{
					msg.message_type=MESSAGE_CHAT_REPLY;
					msg.message_length=msg.data->GetLength();
					mql->send(&msg);
				}
			}
			break;
		case MESSAGE_UPLOAD:
		{
			int upflag=g_config->ReadInt("accept_uploads",1);
			if (upflag&1) {
				C_UploadRequest *r = new C_UploadRequest(message->data);
				if (!stricmp(r->get_dest(),g_client_id_str) || (g_regnick[0] && !stricmp(r->get_dest(),g_regnick)))
				{
					char *fn=r->get_fn();
					log_printf(1,"recv UPLOAD %dB '%s'\n", message->message_length, fn);
					int weirdfn=!strncmp(fn,"..",2) || strstr(fn,":") || strstr(fn,"..\\") || strstr(fn,"../") || fn[0]=='\\' || fn[0] == '/';
					if (!(upflag & 4) || weirdfn)
					{
						char *p=fn;
						while (*p) p++;
						while (p >= fn && *p != '/' && *p != '\\') p--;
						p++;
						if (p != fn || !weirdfn) fn=p;
						else fn="";
					}

					if (fn[0])
					{              
						if (!(upflag & 2)) 
						{
							char str[64];
							MakeID128Str(r->get_guid(),str);
							sprintf(str+strlen(str),":%d",r->get_idx());
							main_handleUpload(str,fn,r);
						}
						else 
						{
							char *t=strdup(fn);
							r->set_fn(t);
							free(t);
							uploadPrompts.Add(r);
							break;
						}
					}
					else
					{
						debug_printf("got upload request that was invalid\n");
					}
				}
				delete r;
			}
		}
			break;
		default:
			debug_printf("recv unknown message %d %dB\n",
				message->message_type, message->message_length);
			break;
	}
}

void handleWasteURL(char *url)
{
	char tmp[2048];
	safe_strncpy(tmp,url,sizeof(tmp));
	char *in=tmp+6,*out=tmp;
	while (*in)
	{
		if (!strncmp(in,"%20",3)) { in+=3; *out++=' '; }
		else *out++=*in++;
	}
	*out=0;

	if (tmp[0] == '?')
	{
		if (!strnicmp(tmp,"?chat=",6))
		{
//			chat_ShowRoom(tmp+6,2);
		}
		else if (!strnicmp(tmp,"?browse=",8))
		{
//			SendMessage(g_mainwnd,WM_COMMAND,IDC_SEARCH,0);
//			Search_Search(tmp+8);
		}
	}
	else
	{
//		SendMessage(g_mainwnd,WM_COMMAND,IDC_SEARCH,0);
//		Search_Search(tmp);
	}
}

void update_set_port()
{
	debug_printf("update_set_port()");

	g_port = (g_route_traffic && g_config->ReadInt("listen",1)) ? g_config->ReadInt("port",1337) : 0;
	if (g_port) 
	{
		debug_printf("update_set_port g_port = %d\n", g_port);
		if (!g_listen) {
			g_listen = new C_Listen((short)g_port);
		}
	}
}

void doDatabaseRescan()
{
	debug_printf("DB: reinitializing database\n");
	if (g_database && g_database != g_newdatabase && g_newdatabase) 
	{
		debug_printf("DB: deleting temp db\n");
		if (g_newdatabase) {
			delete g_newdatabase;
		}
	}             
	g_newdatabase = new C_FileDB();
	if (g_config->ReadInt("use_extlist",0))
		g_newdatabase->UpdateExtList(g_config->ReadString("extlist",g_def_extlist));
	else 
		g_newdatabase->UpdateExtList("*");

	g_newdatabase->Scan(g_config->ReadString("databasepath",""));
	if (!g_database || !g_database->GetNumFiles())
	{
		debug_printf("DB: making scanning db live\n");
		if (g_database) 
		{
			delete g_database;
		}
		g_database=g_newdatabase;
	}
	debug_printf("doDatabaseRescan() finished\n");
}

void main_onGotChannel(char *cnl)
{
	debug_printf("main_onGotChannel - cnl = %s\n", cnl);
	[[WasteApplication sharedApplication] gotChannel: [NSString stringWithCString: cnl]];
}

void main_onGotNick(char *nick, int del)
{
	debug_printf("main_onGotNick - nick = %s\n", nick);
	[[[WasteApplication sharedApplication] clientController] gotNick: [NSString stringWithCString: nick]];
}

void sighandler(int sig)
{
	if (sig == SIGHUP)
	{
		g_do_log=0;
	}
	if (sig == SIGPIPE) debug_printf("got SIGPIPE!\n");
	if (sig == SIGINT)
	{
		g_exit=1;
		debug_printf("got SIGINT, setting global exit flag!\n");
	}
}

void main_AddNodes()
{
	char str[1024+8];
	sprintf(str,"%s.pr5",g_config_prefix);

	FILE *fp=fopen(str,"rb");
	if (fp)
	{
		while (1)
		{
			char line[512];
			fgets(line,512,fp);
			if (feof(fp)) break;
			if (line[0] == '#') continue;
			if (line[strlen(line)-1] == '\n') line[strlen(line)-1]=0;
			if (strlen(line)>1)
			{
				char *p=line;
				while (*p && *p != ':') p++;
				if (*p) *p++=0;
				char *ratingp=p;
				while (*ratingp && *ratingp != ':') ratingp++;
				if (*ratingp) *ratingp++=0;
				int port = atoi(p);
				debug_printf("Adding node: %s:%d\n",line,port);

				// Drop it on the connection list
				AddConnection(line, port, 1);
			}
			else break;
		}
		fclose(fp);
	}
}

void main_SaveNodes()
{
	char str[1024+8];
	sprintf(str,"%s.pr5",g_config_prefix);

	FILE *fp=fopen(str,"wb");
	if (fp)
	{
		int x;
		for (x = 0; x < g_new_net.GetSize(); x ++)
		{
			C_Connection * c = g_new_net.Get(x);
			fprintf(fp, "%s:%d:100\n", c->get_host(), c->get_remote_port());
		}
		fclose(fp);
	}
}

int waste_main(id theController)
{
	int retval;

//	signal(SIGHUP,sighandler);
//	signal(SIGPIPE,sighandler);
//	signal(SIGINT,sighandler);

	fprintf(stderr, "%s starting up...\n",g_nameverstr);

	sprintf(g_scan_status_buf, "server started");
	[[[WasteApplication sharedApplication] clientController] updateStatus];

	strcpy(g_config_prefix, [[NSString stringWithFormat: @"%@/Resources", [[NSBundle mainBundle] resourcePath]] cString]);

	char *p=g_config_prefix;
	while (*p) p++;
	while (p >= g_config_prefix && *p != '/') p--;
	*++p=0;

	strcat(g_config_prefix,"Default");

	MYSRAND();

	char tmp[1024+8];
	sprintf(tmp,"%s.pr0",g_config_prefix);
	NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: tmp] ofType: nil];
	if (!g_config) {
		printf("loading config from %s\n", [path cString]?[path cString]:tmp);
		g_config = new C_Config((char *)([path cString]?[path cString]:tmp));
		g_config->Flush();
	}

	g_do_log=g_config->ReadInt("debuglog",1);

	if (!g_config->ReadString("downloadpath","")[0])
	{
		/* all this code needs to be replaced with NSBundle code */
		strcpy(tmp,"");
		// Holy shit, not again. Fucking pointer code sucks ass.
		char *p=tmp;
		while (*p) p++;
		while (p >= tmp && *p != '/') p--;
		strcpy(++p,"downloads");
		CreateDirectory(tmp,NULL);
		g_config->WriteString("downloadpath",tmp);
	}

	retval = loadPKList();

	debug_printf("loadPKList() loaded %d public keys\n", retval);

	strncpy(g_client_id_str,g_config->ReadString("clientid128",""),32);
	g_client_id_str[32]=0;

	if (MakeID128FromStr(g_client_id_str,&g_client_id)) 
	{
		CreateID128(&g_client_id);
		MakeID128Str(&g_client_id,g_client_id_str);
		g_config->WriteString("clientid128",g_client_id_str);
		g_config->Flush();
		NSCAssert(strlen(g_client_id_str), @"Failed to create client id");
		g_has_started = 0;
		sprintf(g_scan_status_buf, "server stopped");
		[[[WasteApplication sharedApplication] clientController] updateStatus];
		[[[WasteApplication sharedApplication] clientController] stopSpinner];
		return 1;
	}

	if (!g_key.bits) reloadKey(g_config->ReadInt("storepass",0)?
							g_config->ReadString("keypass",NULL):
							NULL);

	if (!g_key.bits) {
		debug_printf("g_key.bits should not be NULL\n");
		return(1);
	}

	if (g_config->ReadInt("db_save",1)) 
	{
		if (!g_newdatabase || !g_database) {
			g_newdatabase=g_database=new C_FileDB();
		}
		if (g_config->ReadInt("use_extlist",0))
			g_newdatabase->UpdateExtList(g_config->ReadString("extlist",g_def_extlist));
		else 
			g_newdatabase->UpdateExtList("*");
		char dbfile[1024+8];
		sprintf(dbfile,"%s.pr2",g_config_prefix);
		NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: dbfile] ofType: nil];
		printf("reading database from %s\n", [path cString]?[path cString]:dbfile);
		g_newdatabase->readIn((char *)([path cString]?[path cString]:dbfile));
		g_scanloadhack=1;
	}

	if (g_config->ReadInt("scanonstartup",1)) doDatabaseRescan();

	{
		char *buf=g_config->ReadString("networkname","");
		if (buf[0])
		{
			SHAify m;
			m.add((unsigned char *)buf,strlen(buf));
			m.final(g_networkhash);
			m.reset();
			g_use_networkhash=1;
		}
		else
		{
			memset(g_networkhash,0,sizeof(g_networkhash));
			g_use_networkhash=0;
		}
	}
	g_conspeed=g_config->ReadInt("conspeed",128);
	g_route_traffic=g_config->ReadInt("route",g_conspeed>64);
	g_forceip=g_config->ReadInt("forceip",0);
	char *addrstr=g_config->ReadString("forceip_addr","");
	if (!strchr(addrstr,'.'))
		g_forceip_addr=atoi(addrstr); /* original integer format */
	else {
		struct in_addr in;
		inet_aton(addrstr,&in); /* numbers-and-dots format */
		g_forceip_addr=in.s_addr;
	}
	g_appendprofiletitles=g_config->ReadInt("appendpt",0);
	g_extrainf=g_config->ReadInt("extrainf",0);
	g_keepup=g_config->ReadInt("keepupnet",4);
	g_use_accesslist=g_config->ReadInt("ac_use",0);
	g_accept_downloads=g_config->ReadInt("downloadflags",7);
	g_do_autorefresh=g_config->ReadInt("dorefresh",0);
	g_max_simul_dl=g_config->ReadInt("recv_maxdl",4);
	g_max_simul_dl_host=g_config->ReadInt("recv_maxdl_host",1);
	g_chat_timestamp=g_config->ReadInt("chat_timestamp",0);
	g_search_showfull=g_config->ReadInt("search_showfull",1);
	g_keydist_flags=g_config->ReadInt("keydistflags",4|2|1);

	g_throttle_flag=g_config->ReadInt("throttleflag",0);
	g_throttle_send=g_config->ReadInt("throttlesend",64);
	g_throttle_recv=g_config->ReadInt("throttlerecv",64);

	if (!g_dns) {
		g_dns = new C_AsyncDNS;
	}

	if (!g_mql) {
		g_mql = new C_MessageQueueList(main_MsgCallback,6);
	}

	updateACList();
	update_set_port();

	safe_strncpy(g_regnick,g_config->ReadString("nick",""),sizeof(g_regnick));
	if (g_regnick[0] == '#' || g_regnick[0] == '&') g_regnick[0]=0;

	char *pubkeyfn="pubkey.txt";
	/* generate and save PUBLIC key */
	char *buf = genMyPubKey(g_regnick, &g_key);
	FILE *fp = fopen(pubkeyfn, "w");
	if (fp == 0)
		printf("Can't open '%s' for writing.\n", pubkeyfn);
	else {
		fwrite(buf, 1, strlen(buf), fp);
		fclose(fp);
		fp = 0;
		debug_printf("Wrote PUBLIC key to '%s'\n", pubkeyfn);
	}
	free(buf);

	signal(SIGHUP,sighandler);
	signal(SIGPIPE,sighandler);
	signal(SIGINT,sighandler);

	// start primary run loop

	NS_DURING

	while (!g_exit)
	{

		g_scanworking=0;
		if (g_scanloadhack)
		{
            g_scanloadhack=0;
            if (g_database)
            {
				sprintf(g_scan_status_buf,"%d files (cached)",g_database->GetNumFiles());
				debug_printf("waste_main() - g_scan_status_buf = %s\n", g_scan_status_buf);
            }
		}
		if (g_newdatabase) 
		{
            int r=g_newdatabase->DoScan(15, (g_newdatabase != g_database && g_database->GetNumFiles()) ? g_database : NULL);
			if (!g_scanworking)
				[[[WasteApplication sharedApplication] clientController] startSpinner];
            if (r != -1)
            {
				g_scanworking=1;
				sprintf(g_scan_status_buf,"Scanning DB item %d",r);
				[[[WasteApplication sharedApplication] clientController] updateStatus];
            }
            else
            {
				g_scanworking=0;
//				sprintf(g_scan_status_buf,"Scanned %d",g_newdatabase->GetNumFiles());
				[[[WasteApplication sharedApplication] clientController] stopSpinner];
				strcpy(g_scan_status_buf, "Server Started");
				[[[WasteApplication sharedApplication] clientController] updateStatus];
				if (g_newdatabase != g_database) 
				{
					debug_printf("DB: replacing old database with temp database\r\n");
					if (g_database) {
//						delete g_database;
					}
					g_database=g_newdatabase;
				}
				else
					debug_printf("DB: removing temp database reference, scanning done\r\n");
				debug_printf("DB: serving %d files\n", g_database->GetNumFiles());
				g_newdatabase=0;
				sprintf(g_scan_status_buf, "DB: serving %d files", g_database->GetNumFiles());
				[[[WasteApplication sharedApplication] clientController] updateStatus];
            }
		}

		int lastqueues=g_mql->GetNumQueues();

		NetKern_Run();

		Xfer_Run();

		[[[WasteApplication sharedApplication] clientController] updateServerStatus];

		if (!g_scanworking && !g_nodes_added) {
			main_AddNodes();
			g_nodes_added = 1;
		}

		g_mql->run(g_route_traffic);

		int newl=g_mql->GetNumQueues();

		if (lastqueues != newl) 
		{
            MYSRANDUPDATE((unsigned char *)&lastqueues,sizeof(lastqueues));
			char text[32];
			if (newl) sprintf(text,"Connections: %d",newl);
			else strcpy(text,"Connect");
			printf("%s\n", text);
//			SetDlgItemText(g_mainwnd,IDC_NETSETTINGS,text);
		}

		if (time(NULL) > g_next_refreshtime && g_do_autorefresh && !g_newdatabase)
		{
            if (!g_scanworking)
            {
				[[[WasteApplication sharedApplication] clientController] stopSpinner];
				if (g_next_refreshtime)
				{
					g_next_refreshtime=0;
					debug_printf("DB: initiating autorescan of files\n");
					if (g_newdatabase) {
//						delete g_newdatabase;
					}
					g_newdatabase = new C_FileDB();
					if (g_config->ReadInt("use_extlist",0))
						g_newdatabase->UpdateExtList(g_config->ReadString("extlist",g_def_extlist));
					else 
						g_newdatabase->UpdateExtList("*");
					g_newdatabase->Scan(g_config->ReadString("databasepath",""));
					if (!g_database->GetNumFiles())
					{
						debug_printf("DB: making scanning db live\n");
						if (g_database) {
//							delete g_database;
						}
						g_database=g_newdatabase;
					}
				}
				else g_next_refreshtime = time(NULL)+60*g_config->ReadInt("refreshint",300);
            }
		}

		if (time(NULL) > g_last_pingtime && newl) // send ping every 2m
		{
			[[[WasteApplication sharedApplication] clientController] startSpinner];
            g_last_pingtime = time(NULL)+120;
            DoPing(NULL);

            // flush config and netq
            g_config->Flush();
			SaveNetQ();
			[[[WasteApplication sharedApplication] clientController] stopSpinner];
		}

		if (time(NULL) > g_last_bpstime && newl) {
			int ms = g_config->ReadInt("bpssec",0);
			if (ms) {
				g_last_bpstime = time(NULL)+ms;
				printf("%dB/s out, %dB/s in, %d connections\n",
		   g_mql->get_total_send_bps(), g_mql->get_total_recv_bps(), newl);
			}
		}

		if (g_config->ReadInt("consumebandwidth",0)) {
			/*pgar - artificially boost bandwidth consumption */
//			DoPing(NULL);
//			main_BroadcastPublicKey(NULL);
		}

		Sleep(33);
	}

	NS_HANDLER
		NSLog(@"An exception occurred : %@", [localException description]);
	NS_ENDHANDLER
	
	// exiting the main run loop

	debug_printf("waste_main() - Shutting down...\n");

	[[[WasteApplication sharedApplication] clientController] startSpinner];

	if (g_config->ReadInt("db_save",1) && g_database && g_database != g_newdatabase) 
	{
		sprintf(g_scan_status_buf, "flushing db");
		[[[WasteApplication sharedApplication] clientController] updateStatus];
		char dbfile[1024+8];
		sprintf(dbfile,"%s.pr2",g_config_prefix);
		NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: dbfile] ofType: nil];
		printf("writing database to %s\n", [path cString]?[path cString]:dbfile);
		g_database->writeOut((char *)([path cString]?[path cString]:dbfile));
	}

	main_SaveNodes();

	msgStats();

	// Since we're not actually quiting the app, just this thread, don't get rid of these objects

	if (g_listen) {
		delete g_listen;
	}
	if (g_dns) {
		delete g_dns;
	}
	if (g_newdatabase != g_database) {
		if (g_newdatabase) {
			delete g_newdatabase;
		}
	}
	if (g_database) {
		delete g_database;
	}
	int x;
	for (x = 0; x < g_recvs.GetSize(); x ++) if (g_recvs.Get(x)) delete g_recvs.Get(x);
	for (x = 0; x < g_sends.GetSize(); x ++) if (g_sends.Get(x)) delete g_sends.Get(x);
	for (x = 0; x < g_new_net.GetSize(); x ++) if (g_new_net.Get(x)) delete g_new_net.Get(x);
	for (x = 0; x < g_uploads.GetSize(); x ++) if (g_uploads.Get(x)) free(g_uploads.Get(x));

	if (g_mql) {
		delete g_mql;
	}
	if (g_config) {
		delete g_config;
	}

	g_has_started = 0;

	sprintf(g_scan_status_buf, "server stopped");

	[[[WasteApplication sharedApplication] clientController] updateStatus];
	[[[WasteApplication sharedApplication] clientController] stopSpinner];

	[NSApp replyToApplicationShouldTerminate: YES];

	return 0;
}

void stop_waste(void)
{
	g_exit = 1;
	debug_printf("g_exit is now %d\n", g_exit);
}

@end
