#import "PrefsController.h"

#import "WasteApplication.h"

#include "util.h"
#include "prefs.h"

#include <unistd.h>

extern int loadPKList(char *fn); // returns num of keys loaded
extern void savePKList();
extern void debug_printf(char *, ...);

extern char g_config_prefix[1024];

extern void copyMyPubKey();

@implementation PrefsController

- (void)awakeFromNib;
{
	[publicKeyTableView setDelegate: self];
	[publicKeyTableView setDataSource: self];
	[publicKeyTableView setAutosaveTableColumns: YES];
	[publicKeyTableView setAutosaveName: @"pubkeytableview"];
	[publicKeyTableView setAutoresizesAllColumnsToFit: NO];
	[publicKeyTableView setAutosaveTableColumns: YES];
	
	[pendingKeyTableView setDelegate: self];
	[pendingKeyTableView setDataSource: self];
	[pendingKeyTableView setAutosaveTableColumns: YES];
	[pendingKeyTableView setAutosaveName: @"pendingtableview"];
	[pendingKeyTableView setAutoresizesAllColumnsToFit: NO];
	[pendingKeyTableView setAutosaveTableColumns: YES];
	
	[ipAddressTableView setDelegate: self];
	[ipAddressTableView setDataSource: self];
	[ipAddressTableView setAutosaveTableColumns: YES];
	[ipAddressTableView setAutosaveName: @"ipaddresstableview"];
	[ipAddressTableView setAutoresizesAllColumnsToFit: NO];
	[ipAddressTableView setAutosaveTableColumns: YES];
}

- (IBAction)close:(id)sender
{
	[prefsPanel performClose: nil];
}

- (IBAction) copyPublicKey: (id)sender;
{
	copyMyPubKey();
}

- (IBAction)showPrefs:(id)sender
{
	[prefsPanel makeKeyAndOrderFront: nil];
}

- (IBAction)importPublicKeys:(id)sender;
{
	NSOpenPanel * panel = [NSOpenPanel openPanel];
	NSString * path = NSHomeDirectory();
	NSArray * types = [NSArray arrayWithObjects: @"pub", @"txt", @"pr3", @"key", nil];
	[panel setAllowsMultipleSelection: NO];
	[panel setCanChooseDirectories: NO];
	[panel setCanChooseFiles: YES];
	[panel setAccessoryView: [keygenAuxView retain]];
	[panel beginSheetForDirectory: path file: nil types: types modalForWindow: prefsPanel modalDelegate: self didEndSelector: (SEL)@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo: sender];
}

- (void)importPublicKeysFromFile: (NSString *)file;
{
	debug_printf("importPublicKeysFromFile: %s\n", [file cString]);
	if ([file length])
	{
		int num = loadPKList((char *)[file cString]);
		if (num)
		{
			savePKList();
			debug_printf("Imported %d public keys successfully.\n",num);
			NSRunAlertPanel(@"Public Key Import", [NSString stringWithFormat: @"%d keys loaded successfully", num], @"OK", nil, nil);
		}
		else
		{
			debug_printf("No public keys found in file '%s'\n", [file cString]);
		}
	}
	[publicKeyTableView reloadData];
}

- (IBAction)importKeyText:(id)sender;
{
	char * TEMPKEY = "/tmp/_waste_imported.key";
	NSString * keyText = [keygenImportKeyTextField stringValue];
//	debug_printf("importKeyText: keyText = %s\n", [keyText cString]);
//	[NSApp stopModal];
	int num = 0;
	FILE *fp = fopen(TEMPKEY, "wt");
	if (fp) {
		const char * p = [keyText cString];
		fwrite(p, 1, strlen(p), fp);
		fclose(fp);
		num = loadPKList(TEMPKEY);
		unlink(TEMPKEY);
	}
	if (num) {
		NSRunAlertPanel(@"Public Key Import", [NSString stringWithFormat: @"%d keys imported successfully\n(Click Cancel if you're done)", num], @"OK", nil, nil);
	}
}

- (IBAction) viewPublicKeys: (id)sender;
{
	char str[2048];
	NSAssert(g_config_prefix && strlen(g_config_prefix), @"g_config_prefix should not be empty");
	sprintf(str,"%s.pr3",g_config_prefix);
	NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: str] ofType: nil];
	debug_printf("opening public key list file '%s'\n", (char *)([path cString]?[path cString]:str));
	if (strlen([path cString]?[path cString]:str)) {
		NSString *publicKeyPath = [NSString stringWithCString: (char *)([path cString]?[path cString]:str)];
		[[NSWorkspace sharedWorkspace] openFile: publicKeyPath withApplication: @"TextEdit" andDeactivate: YES];
	}
}

- (void)openPanelDidEnd: (NSOpenPanel *)sheet returnCode: (int)returnCode contextInfo: (void  *)contextInfo;
{
	if (publicKeyImportPublicKeyButton == contextInfo) {
		if (returnCode == NSAlertDefaultReturn) {
			if ([[sheet filenames] count]) {
				[self importPublicKeysFromFile: [[sheet filenames] objectAtIndex: 0]];
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// NSTableView datasource methods
//////////////////////////////////////////////////////////////////////////////

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
#ifdef _DEBUG_
//	NSLog(@"%@ - numberOfRowsInTableView: %@", [self class], aTableView);
#endif
	if ([aTableView tag] == 0) {
//		NSAssert(g_pklist != NULL, @"publicKeys array should not be nil");
		return g_pklist.GetSize();
	}
	if ([aTableView tag] == 1) {
//		NSAssert(pendingKeys, @"pendingKeys array should not be nil");
		return g_pklist_pending.GetSize();
	}
	if ([aTableView tag] == 2) {
//		NSAssert(ipAddresses, @"ipAddresses array should not be nil");
//		return [ipAddresses count];
	}
	return 0;
}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)row
{
#ifdef _DEBUG_
//	NSLog(@"%@ - tableView: %@ objectValueForTableColumn: %@ row: %d", [self class], aTableView, aTableColumn, row);
#endif
	if (aTableView == publicKeyTableView) {
		PKitem *p=g_pklist.Get(row);
		if ([[aTableColumn identifier] intValue] == 0) {
			NSAssert(p, @"p should not be nil");
			return [NSString stringWithCString: p->name];
		}
		if ([[aTableColumn identifier] intValue] == 1) {
			char buf[128];
			int a;
			for (a = 0; a < SHA_OUTSIZE; a ++) sprintf(buf+a*2,"%02X",p->hash[a]);
			return [NSString stringWithCString: buf];
		}
	}
	if (aTableView == pendingKeyTableView) {
		if ([[aTableColumn identifier] intValue] == 1) {
	//		NSAssert(pendingKeys != nil, @"pendingKeys array should not be nil");
	//		NSString *key = [pendingKeys objectAtIndex: row];
	//		NSAssert(key != nil, @"key string should not be nil");
	//		return key;
	//		[cell setStringValue: key];
	//		if ([cell image] == nil) {
	//			[cell setImage: [NSImage imageNamed: @"allusers.ico"]];
	//		}
		}
	}

	if (aTableView == ipAddressTableView) {
		if ([[aTableColumn identifier] intValue] == 2) {
	//		NSAssert(ipAddresses != nil, @"ipAddresses array should not be nil");
	//		NSString *ip = [ipAddresses objectAtIndex: row];
	//		NSAssert(ip != nil, @"channel string should not be nil");
	//		return ip;
	//		[cell setStringValue: ip];
	//		if ([cell image] == nil) {
	//			[cell setImage: [NSImage imageNamed: @"allusers.ico"]];
	//		}
		}
	}
	return @"";
}

//////////////////////////////////////////////////////////////////////////////
// NSTabView delegate methods
//////////////////////////////////////////////////////////////////////////////

- (BOOL)tabView:(NSTabView *)tabView shouldSelectTabViewItem:(NSTabViewItem *)tabViewItem;
{
	if (![(NSString *)[tabViewItem identifier] length]) {
		return NO;
	}
	switch ([[tabViewItem identifier] intValue]) {
		case 0: [publicKeyTableView reloadData]; break;
		case 1: [pendingKeyTableView reloadData]; break;
		case 2: [ipAddressTableView reloadData]; break;
	}
	return YES;
}

@end
