/*
 WASTE - keygen.cpp (Keypair generation UI)
 Copyright (C) 2003 Nullsoft, Inc.

 WASTE is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 WASTE  is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with WASTE; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#import "SetupController.h"

#import "WasteApplication.h"

#include <time.h>
#include <stdio.h>
#include "resource.h"

#include "main.h"
extern "C" {
#include "rsa/r_random.h"
};

#include "platform.h"

#include "listen.h"
#include "mqueuelist.h"
#include "filedb.h"
#include "config.h"

#ifdef _WIN32
#include <windows.h>
#endif

#import <Cocoa/Cocoa.h>

extern void debug_printf(char *, ...);

static char kg_privout[1024];
static char kg_passbf[SHA_OUTSIZE];

static int kg_keysize;
static R_RANDOM_STRUCT kg_random;
static int kg_going;

static time_t kg_start_time;
static unsigned int kg_movebuf[7];
static int kg_movebuf_cnt;

#ifdef _WIN32
static WNDPROC kg_oldWndProc;
#endif

#ifdef _WIN32
static HANDLE kg_thread;
#endif

static R_RSA_PRIVATE_KEY kg_privKey;

#ifdef _WIN32
static HWND kg_hwndDlg;
#endif

void copyMyPubKey();

static void writeBFdata(FILE *out, BLOWFISH_CTX *ctx, void *data, unsigned int len, UINT4 *cbcl, UINT4 *cbcr, int *lc)
{
	unsigned int x;
	UINT4 *p=(UINT4 *)data;
#ifdef _DEBUG_
//	id tmpdata = [NSData dataWithBytes: (unsigned char *)data length: len];
//	id string = [[NSMutableString alloc] initWithData: tmpdata encoding: [NSString defaultCStringEncoding]];
//	NSLog(@"writeBFdata() - data = %@", tmpdata);
#endif
	debug_printf("SetupController.mm : writeBFdata()\n");
	for (x = 0; x < len; x += 8)
	{
		UINT4 pp[2];
		pp[0]=p[0];
		pp[1]=p[1];

		pp[0]^=*cbcl;
		pp[1]^=*cbcr;
		Blowfish_Encrypt(ctx,pp,pp+1);
		*cbcl=pp[0];
		*cbcr=pp[1];
		int c;
		for (c = 0; c < 8; c ++)
		{
			fprintf(out,"%02X",((unsigned char *)pp)[c]);
			printf("%02X",((unsigned char *)pp)[c]);
			if (++*lc % 30 == 0) fprintf(out,"\n");
			if (*lc % 4 == 0) printf(" ");
			if (*lc % 16 == 0) printf("\n");
		}
//#ifdef _DEBUG_
//	id tmpdata = [NSData dataWithBytes: ((unsigned char *)pp) length: 8];
//	id string = [[NSMutableString alloc] initWithData: tmpdata encoding: [NSString defaultCStringEncoding]];
//	NSLog(@"writeBFdata() - pp = %@", tmpdata);
//#endif
		p+=2;
	}
}

int kg_writePrivateKey(char *fn, R_RSA_PRIVATE_KEY *key, R_RANDOM_STRUCT *rnd, char *passhash)
{
	FILE *fp;
	//	int ks=(key->bits+7)/8;
	int x;
	int lc=8;

	fp=fopen(fn,"wt");
	if (!fp) return 1;

	fprintf(fp,"WASTE_PRIVATE_KEY 10 %d\n",key->bits);

	UINT4 tl[2];
	R_GenerateBytes((unsigned char *)&tl,8,rnd);
	for (x = 0; x < 8; x ++)
	{
//#if BYTE_ORDER == BIG_ENDIAN
#ifdef __BIG_ENDIAN__
		fprintf(fp,"%02X",(int)((tl[x/4]>>((3-(x&3))*8))&0xff));
#else
		fprintf(fp,"%02X",(int)((tl[x/4]>>((x&3)*8))&0xff));
#endif
	}
	BLOWFISH_CTX c;
	Blowfish_Init(&c,(unsigned char *)passhash,SHA_OUTSIZE);
	char buf[9]="PASSWORD";

	writeBFdata(fp,&c,buf,8,tl,tl+1,&lc);
	writeBFdata(fp,&c,key->modulus,sizeof(key->modulus),tl,tl+1,&lc);
	writeBFdata(fp,&c,key->publicExponent,sizeof(key->publicExponent),tl,tl+1,&lc);
	writeBFdata(fp,&c,key->exponent,sizeof(key->exponent),tl,tl+1,&lc);
	writeBFdata(fp,&c,key->prime,sizeof(key->prime),tl,tl+1,&lc);
	writeBFdata(fp,&c,key->primeExponent,sizeof(key->primeExponent),tl,tl+1,&lc);
	writeBFdata(fp,&c,key->coefficient,sizeof(key->coefficient),tl,tl+1,&lc);

	if (lc % 30) fprintf(fp,"\n");
	fprintf(fp,"WASTE_PRIVATE_KEY_END\n");
	fclose(fp);

	memset(&c,0,sizeof(c));//safety

	return 0;
}

int keygen(NSLock * keygenLock)
{
	[keygenLock lock];
	kg_start_time=time(NULL);
	R_RSA_PROTO_KEY protoKey;
	protoKey.bits=kg_keysize;
	protoKey.useFermat4=1;
	R_RSA_PUBLIC_KEY kg_pubKey;
	memset(&kg_pubKey,0,sizeof(kg_pubKey));
	if (R_GeneratePEMKeys(&kg_pubKey,&kg_privKey,&protoKey,&kg_random) ||
	 kg_writePrivateKey(kg_privout,&kg_privKey,&kg_random,kg_passbf))
	{
		fprintf(stderr, "Error creating keys\n");
	}
	else
	{
		kg_start_time=time(NULL)-kg_start_time;
	}
	memset(kg_passbf,0,SHA_OUTSIZE);
	memset(&kg_pubKey,0,sizeof(kg_pubKey));
	memset(&kg_random,0,sizeof(kg_random));

	memcpy(&g_key,&kg_privKey,sizeof(R_RSA_PRIVATE_KEY));
	SHAify m;
	m.add((unsigned char *)g_key.modulus,MAX_RSA_MODULUS_LEN);
	m.add((unsigned char *)g_key.publicExponent,MAX_RSA_MODULUS_LEN);
	m.final(g_pubkeyhash);

	memset(&kg_privKey,0,sizeof(R_RSA_PRIVATE_KEY));
	[keygenLock unlock];
	return 0;
}

#ifdef _WIN32
static DWORD WINAPI keyThread(LPVOID p)
#endif
int keyThread()
{
	kg_start_time=time(NULL);
	R_RSA_PROTO_KEY protoKey;
	protoKey.bits=kg_keysize;
	protoKey.useFermat4=1;
	R_RSA_PUBLIC_KEY kg_pubKey;
	memset(&kg_pubKey,0,sizeof(kg_pubKey));
	if (R_GeneratePEMKeys(&kg_pubKey,&kg_privKey,&protoKey,&kg_random) ||
	 kg_writePrivateKey(kg_privout,&kg_privKey,&kg_random,kg_passbf))
	{
#ifdef _WIN32
		SetDlgItemText(kg_hwndDlg,IDC_LABEL1,"Error creating keys");
		SetDlgItemText(kg_hwndDlg,IDC_LABEL5,"");
		MessageBox(kg_hwndDlg,"Error calling GeneratePEMKeys()",APP_NAME " Key Generation Error", MB_OK|MB_ICONSTOP);
#endif
	}
	else
	{
		kg_start_time=time(NULL)-kg_start_time;
#ifdef _WIN32
		// "adjust UI" PostMessage(kg_hwndDlg,WM_USER+0x131,0,1);
#endif
	}
	memset(kg_passbf,0,SHA_OUTSIZE);
	memset(&kg_pubKey,0,sizeof(kg_pubKey));
	memset(&kg_random,0,sizeof(kg_random));

	memcpy(&g_key,&kg_privKey,sizeof(R_RSA_PRIVATE_KEY));
	SHAify m;
	m.add((unsigned char *)g_key.modulus,MAX_RSA_MODULUS_LEN);
	m.add((unsigned char *)g_key.publicExponent,MAX_RSA_MODULUS_LEN);
	m.final(g_pubkeyhash);

	memset(&kg_privKey,0,sizeof(R_RSA_PRIVATE_KEY));

#ifdef _WIN32
	KillTimer(kg_hwndDlg,1);
#endif
	return 0;
}

#ifdef _WIN32
static BOOL CALLBACK kg_newWndProc(HWND hwndDlg, UINT uMsg, WPARAM wParam,LPARAM lParam)
{
	if (uMsg == WM_MOUSEMOVE && kg_going==1)
	{
		kg_movebuf[kg_movebuf_cnt%7]+=lParam;
		kg_movebuf[(kg_movebuf_cnt+1)%7]+=GetTickCount();
		kg_movebuf[(kg_movebuf_cnt+2)%7]+=GetMessageTime()+GetMessagePos();
		if (++kg_movebuf_cnt >= 29)
		{
			kg_movebuf_cnt=0;
			R_RandomUpdate(&kg_random,(unsigned char *)kg_movebuf,sizeof(kg_movebuf));

			unsigned int bytesNeeded;
			R_GetRandomBytesNeeded(&bytesNeeded, &kg_random);
			SendDlgItemMessage(hwndDlg,IDC_PROGRESS1,PBM_SETPOS,(WPARAM)(64-bytesNeeded/4),0);
			if (!bytesNeeded)
			{
				kg_going=2;
				SendDlgItemMessage(hwndDlg,IDC_PROGRESS1,PBM_SETPOS,0,0);
				SetDlgItemText(hwndDlg,IDC_LABEL1,"Generating key pair... please wait");
				char buf[128];
				wsprintf(buf,"(this may take as long as %d minutes)",kg_keysize<=1024?1:kg_keysize<=2048?3:kg_keysize<=3072?10:20);
				SetDlgItemText(hwndDlg,IDC_LABEL5,buf);
				SetTimer(hwndDlg,1,250,NULL);
				DWORD id;
				kg_thread=CreateThread(NULL,0,keyThread,NULL,0,&id);
			}
		}
	}
	return CallWindowProc(kg_oldWndProc,hwndDlg,uMsg,wParam,lParam);
}
#endif

#ifdef _WIN32
static BOOL WINAPI dlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_USER+0x131: // success, finish writing out
			SendDlgItemMessage(hwndDlg,IDC_PROGRESS1,PBM_SETPOS,(WPARAM)64,0);
			{
				char buf[123];
				wsprintf(buf,"Completed in %d minutes %d seconds",kg_start_time/60,kg_start_time%60);
				SetDlgItemText(hwndDlg,IDC_LABEL1,buf);
				SetDlgItemText(hwndDlg,IDC_LABEL5,"");
				MessageBox(hwndDlg,buf,APP_NAME " Key Generator - Completed",MB_OK|MB_ICONINFORMATION);
			}
				SetDlgItemText(hwndDlg,IDOK,"Close");
			// makeKeyAndOrderFront: EnableWindow(GetDlgItem(hwndDlg,IDOK),1);
			SetActiveWindow(GetDlgItem(hwndDlg,IDOK));

			// makeKeyAndOrderFront: EnableWindow(GetDlgItem(hwndDlg,IDCANCEL),0);
			if (kg_thread)
			{
				CloseHandle(kg_thread);
				kg_thread=0;
			}
				// performClose: nil EndDialog(hwndDlg,1); // yay
			break;
		case WM_INITDIALOG:
			kg_hwndDlg=hwndDlg;
			SetWindowText(hwndDlg,APP_NAME " Key Generator");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_ADDSTRING,0,(LPARAM)"1024 bits (weak, not recommended)");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_ADDSTRING,0,(LPARAM)"1536 bits (recommended)");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_ADDSTRING,0,(LPARAM)"2048 bits (slower, recommended)");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_ADDSTRING,0,(LPARAM)"3072 bits (slow, not recommended)");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_ADDSTRING,0,(LPARAM)"4096 bits (very slow, not recommended)");
			SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_SETCURSEL,1,0);
			break;
		case WM_CLOSE: // performClose: nil EndDialog(hwndDlg,0); break;
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCANCEL:
					if (kg_thread)
					{
						if (WaitForSingleObject(kg_thread,0) == WAIT_TIMEOUT) TerminateThread(kg_thread,0);
						CloseHandle(kg_thread);
						kg_thread=0;
					}
					// performClose: nil EndDialog(hwndDlg,0);
					break;
				case IDOK:
					if (kg_going == 2)
					{
						// performClose: nil EndDialog(hwndDlg,1);
					}
					if (!kg_going)
					{
						int wndhide[]={
							IDC_EDIT1,IDC_EDIT2,
							IDC_LABEL3,IDC_LABEL4,IDC_LABEL6,IDC_COMBO1
						};
						switch (SendDlgItemMessage(hwndDlg,IDC_COMBO1,CB_GETCURSEL,0,0))
						{
							case 0:
								kg_keysize=1024;
								break;
							case 1:
								kg_keysize=1536;
								break;
							case 2:
								kg_keysize=2048;
								break;
							case 3:
								kg_keysize=3072;
								break;
							case 4:
								kg_keysize=4096;
								break;
							default:
								kg_keysize=2048;
						}

						char pass1[1024],pass2[1024];

						// [... stringValue] GetDlgItemText(hwndDlg,IDC_EDIT1,pass1,sizeof(pass1));
						// [... stringValue] GetDlgItemText(hwndDlg,IDC_EDIT2,pass2,sizeof(pass2));
						if (strcmp(pass1,pass2))
						{
							MessageBox(hwndDlg,"Passphrase mistyped",APP_NAME " Key Generator Error",MB_OK|MB_ICONSTOP);
							break;
						}

						SHAify c;
						c.add((unsigned char *)pass1,strlen(pass1));
						c.final((unsigned char *)kg_passbf);
						memset(pass1,0,sizeof(pass1));
						memset(pass2,0,sizeof(pass2));

						int x;
						for (x = 0; x < sizeof(wndhide)/sizeof(wndhide[0]); x ++)
							ShowWindow(GetDlgItem(hwndDlg,wndhide[x]),SW_HIDE);

						ShowWindow(GetDlgItem(hwndDlg,IDC_LABEL5),SW_SHOWNA);
						SetDlgItemText(hwndDlg,IDC_LABEL1,"Giving random number generator more randomness, please move");

						// makeKeyAndOrderFront: EnableWindow(GetDlgItem(hwndDlg,IDOK),0);

						ShowWindow(GetDlgItem(hwndDlg,IDC_PROGRESS1),SW_SHOWNA);
						SendDlgItemMessage(hwndDlg,IDC_PROGRESS1,PBM_SETRANGE,0,MAKELPARAM(0,64));
						kg_oldWndProc=(WNDPROC) SetWindowLong(hwndDlg,GWL_WNDPROC,(LONG)kg_newWndProc);

						kg_going=1;
						R_RandomInit(&kg_random);
						unsigned char buf[16];
						R_GenerateBytes(buf,sizeof(buf),&g_random);
						R_RandomUpdate(&kg_random,buf,sizeof(buf));
						memset(buf,0,sizeof(buf));
					}
					break;
			}
			return 0;
		case WM_DESTROY:
			if (kg_oldWndProc)
				SetWindowLong(hwndDlg,GWL_WNDPROC,(LONG)kg_oldWndProc);
			kg_oldWndProc=0;
			return 0;
		case WM_TIMER:
			if (wParam == 1)
			{
				static int pos;
				if (++pos == 64) pos=0;
				SendDlgItemMessage(hwndDlg,IDC_PROGRESS1,PBM_SETPOS,(WPARAM)pos,0);
			}
			break;
	}
	return 0;
}
#endif

static int running=0;

#ifdef _WIN32
void RunKeyGen(HWND hwndParent, char *keyout)
#endif
void RunKeyGen(char *keyout)
{
	if (!running)
	{
		running=1;
		memset(&kg_privKey,0,sizeof(kg_privKey));
#ifdef _WIN32
		kg_thread=0;
#endif
		kg_start_time=0;
		memset(kg_movebuf,0,sizeof(kg_movebuf));
		kg_movebuf_cnt=0;
#ifdef _WIN32
		kg_oldWndProc=0;
#endif
		kg_going=0;
		kg_keysize=2048;
		memset(&kg_random,0,sizeof(kg_random));
#ifdef _WIN32
		kg_privout=keyout;
		DialogBox(g_hInst,MAKEINTRESOURCE(IDD_KEYGEN),hwndParent,dlgProc);
#endif
		running=0;
	}
}

/*

 void writeBFdata(FILE *out, BLOWFISH_CTX *ctx, void *data, unsigned int len, unsigned int *cbcl, unsigned int *cbcr, int *lc)
 {
#ifdef _DEBUG_
	 id tmpdata = [NSData dataWithBytes: data length: len];
	 id string = [[NSMutableString alloc] initWithData: tmpdata encoding: [NSString defaultCStringEncoding]];
	 NSLog(@"writeBFdata() - data = %@", tmpdata);
#endif
	 unsigned int x;
	 UINT4 *p=(UINT4 *)data;
	 for (x = 0; x < len; x += 8)
	 {
		 unsigned int pp[2];
		 pp[0]=p[0];
		 pp[1]=p[1];

		 pp[0]^=*cbcl;
		 pp[1]^=*cbcr;
		 Blowfish_Encrypt(ctx,pp,pp+1);
		 *cbcl=pp[0];
		 *cbcr=pp[1];
		 int c;
		 for (c = 0; c < 8; c ++)
		 {
			 fprintf(out,"%02X",((unsigned char *)pp)[c]);
			 if (++*lc % 30 == 0) fprintf(out,"\n");
		 }
#ifdef _DEBUG_
		 {
			 id tmpdata = [NSData dataWithBytes: (unsigned char *)pp length: 8];
			 id string = [[NSMutableString alloc] initWithData: tmpdata encoding: [NSString defaultCStringEncoding]];
			 NSLog(@"writeBFdata() - pp = %@", tmpdata);
		 }
#endif
		 p+=2;
	 }
 }

 int kg_writePrivateKey(char *fn, R_RSA_PRIVATE_KEY *key, R_RANDOM_STRUCT *rnd, char *passhash)
 {
#ifdef _DEBUG_
	 NSLog(@"kg_writePrivateKey() - fn = %s", fn);
#endif
	 FILE *fp;
	 int x;
	 int lc=8;
	 NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: fn] ofType: nil];
	 printf("writing private key to %s\n", (char *)([path cString]?[path cString]:fn));

	 fp=fopen([path cString]?[path cString]:fn,"wt");
	 if (!fp) return 1;

	 fprintf(fp,"WASTE_PRIVATE_KEY 10 %d\n",key->bits);

	 unsigned int tl[2];
	 R_GenerateBytes((unsigned char *)&tl,8,rnd);
	 for (x = 0; x < 8; x ++)
	 {
		 fprintf(fp,"%02X",(unsigned int)(tl[x/4]>>((x&3)*8))&0xff);
	 }
	 BLOWFISH_CTX c;
	 Blowfish_Init(&c,(unsigned char *)passhash,SHA_OUTSIZE);
#ifdef _DEBUG_
	 NSData * data = [NSData dataWithBytes: passhash length: SHA_OUTSIZE];
	 NSLog(@"SetupController : kg_writePrivateKey() - passhash = %@", data);
#endif
	 char buf[9]="PASSWORD";
	 writeBFdata(fp,&c,buf,8,tl,tl+1,&lc);

#define WPK(x) writeBFdata(fp,&c,key->x,sizeof(key->x),tl,tl+1,&lc);
	 WPK(modulus)
		 WPK(publicExponent)
		 WPK(exponent)
		 WPK(prime)
		 WPK(primeExponent)
		 WPK(coefficient)
#undef WPK

		 if (lc % 30) fprintf(fp,"\n");
	 fprintf(fp,"WASTE_PRIVATE_KEY_END\n");
	 fclose(fp);

	 memset(&c,0,sizeof(c));	// safety

	 return 0;
 }

 int keyThread()
 {
#ifdef _DEBUG_
	 NSLog(@"keyThread()");
#endif

	 kg_start_time=time(NULL);
	 R_RSA_PROTO_KEY protoKey;
	 protoKey.bits=kg_keysize = 2048;
	 protoKey.useFermat4=1;
	 R_RSA_PUBLIC_KEY kg_pubKey;
	 memset(&kg_pubKey,0,sizeof(kg_pubKey));
	 if (R_GeneratePEMKeys(&kg_pubKey,&kg_privKey,&protoKey,&kg_random) || kg_writePrivateKey(kg_privout,&kg_privKey,&kg_random,kg_passbf))
	 {
#ifdef _DEBUG_
		 NSLog(@"keyThread() - Error creating keys");
#endif
		 kg_going = 0;
		 return 1;		// Error condition flag is non-zero return value
	 }
	 else
	 {
		 kg_start_time=time(NULL)-kg_start_time;
#ifdef _DEBUG_
		 NSLog(@"keyThread() - PostMessage");
#endif
	 }
	 memset(kg_passbf,0,SHA_OUTSIZE);
	 memset(&kg_pubKey,0,sizeof(kg_pubKey));
	 memset(&kg_random,0,sizeof(kg_random));

	 memcpy(&g_key,&kg_privKey,sizeof(R_RSA_PRIVATE_KEY));
	 SHAify m;
	 m.add((unsigned char *)g_key.modulus,MAX_RSA_MODULUS_LEN);
	 m.add((unsigned char *)g_key.publicExponent,MAX_RSA_MODULUS_LEN);
	 m.final(g_pubkeyhash);

	 memset(&kg_privKey,0,sizeof(R_RSA_PRIVATE_KEY));

	 kg_going = 0;
	 return 0;
 }
 */

@implementation SetupController

- (id)init;
{
	self = [super init];
	keygenLock = [[NSLock alloc] init];
	[self setNeedsRandomNumbers: YES];
	[self setShouldQuit: NO];
	return self;
}

- (NSLock *)keygenLock;
{
	return keygenLock;
}

- (void)setShouldQuit: (BOOL)aFlag;
{
	should_quit = aFlag;
}

- (BOOL)shouldQuit;
{
	return should_quit;
}

- (void)setNeedsRandomNumbers: (BOOL)aFlag;
{
	needs_random_numbers = aFlag;
}

- (BOOL)needsRandomNumbers;
{
	return needs_random_numbers;
}

- (void)awakeFromNib;
{
//	if (!windowSizes) {
//		windowSizes = [[NSMutableDictionary dictionary] retain];
//	}
	[identityView retain];
	[passphraseView retain];
	[randomView retain];
	[keygenView retain];
	[profileView retain];
//	if (![windowSizes objectForKey: @"identityView"]) {
//		[windowSizes setObject: [NSString stringWithFormat: @"%f %f %f %f", [identityView bounds].origin.x, [identityView bounds].origin.y, [identityView bounds].size.width, [identityView bounds].size.height] forKey: @"identityView"];
//		[windowSizes setObject: [NSString stringWithFormat: @"%f %f %f %f", [passphraseView bounds].origin.x, [passphraseView bounds].origin.y, [passphraseView bounds].size.width, [passphraseView bounds].size.height] forKey: @"passphraseView"];
//		[windowSizes setObject: [NSString stringWithFormat: @"%f %f %f %f", [randomView bounds].origin.x, [randomView bounds].origin.y, [randomView bounds].size.width, [randomView bounds].size.height] forKey: @"randomView"];
//		[windowSizes setObject: [NSString stringWithFormat: @"%f %f %f %f", [keygenView bounds].origin.x, [keygenView bounds].origin.y, [keygenView bounds].size.width, [keygenView bounds].size.height] forKey: @"keygenView"];
//		[windowSizes setObject: [NSString stringWithFormat: @"%f %f %f %f", [profileView bounds].origin.x, [profileView bounds].origin.y, [profileView bounds].size.width, [profileView bounds].size.height] forKey: @"profileView"];
//	}
//	[self showIdentity: nil];
}

- (IBAction)cancel:(id)sender
{
//	[NSApp stopModal];
	[self setNeedsRandomNumbers: YES];
	[self setShouldQuit: YES];
	[self showPassphrase: nil];
}

- (IBAction)quit:(id)sender
{
//	[NSApp stopModal];
	[mainWindow close];
	[[WasteApplication sharedApplication] setIsSettingUp: NO];
//	[NSApp terminate: self];
}

- (IBAction)run:(id)sender
{
	if (![self validateProfile]) {
		return;
	}
//	[NSApp stopModal];
	[mainWindow close];
	[[WasteApplication sharedApplication] start: nil];
}

- (IBAction)runKeyGenerator:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - runKeyGenerator: %@", [self class], sender);
#endif
}

- (IBAction)importKeyFromFile:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - importKeyFromFile: %@", [self class], sender);
#endif
}

- (IBAction)copyPublicKeyToClipboard:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - copyPublicKeyToClipboard: %@", [self class], sender);
#endif
	copyMyPubKey();
}

- (IBAction)importPublicKeys:(id)sender;
{
	NSOpenPanel * panel = [NSOpenPanel openPanel];
	NSString * path = NSHomeDirectory();
	NSArray * types = [NSArray arrayWithObjects: @"pub", @"txt", @"pr3", @"key", nil];
	[panel setAllowsMultipleSelection: NO];
	[panel setCanChooseDirectories: NO];
	[panel setCanChooseFiles: YES];
	[panel setAccessoryView: keygenAuxView];
	[panel beginSheetForDirectory: path file: nil types: types modalForWindow: mainWindow modalDelegate: self didEndSelector: (SEL)@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo: sender];
}

- (void)importPublicKeysFromFile: (NSString *)file;
{
	debug_printf("importPublicKeysFromFile: %s\n", [file cString]);
	if ([file length])
	{
		int num = loadPKList((char *)[file cString]);
		if (num)
		{
			savePKList();
			[keysLoadedLabel setStringValue: [NSString stringWithFormat: @"%d keys loaded total", num]];
			NSRunAlertPanel(@"Public Key Import", [NSString stringWithFormat: @"%d keys loaded successfully", num], @"OK", nil, nil);
			debug_printf("Imported %d public keys successfully.", num);
		}
		else
		{
			debug_printf("No public keys found in file '%s'", [file cString]);
		}
	}
}

- (IBAction)importKeyText:(id)sender;
{
#define TEMPKEY "/tmp/_waste_imported.key"
//	[NSApp stopModal];
	int num = 0;
	FILE *fp = fopen(TEMPKEY, "wt");
	if (fp) {
		const char * p = [[keygenImportKeyTextField stringValue] cString];
		fwrite(p, 1, strlen(p), fp);
		fclose(fp);
		num = loadPKList(TEMPKEY);
		unlink(TEMPKEY);
	}
#undef TEMPKEY
	if (num) {
		NSRunAlertPanel(@"Public Key Import", [NSString stringWithFormat: @"%d keys imported successfully\n(Click Cancel if you're done)", num], @"OK", nil, nil);
	}
}

- (IBAction)setDownloadDirectory:(id)sender;
{
	NSOpenPanel * panel = [NSOpenPanel openPanel];
	NSString * path = NSHomeDirectory();
	[panel setAllowsMultipleSelection: YES];
	[panel setCanChooseDirectories: YES];
	[panel setCanChooseFiles: NO];
	//	[NSApp stopModal];
	[panel beginSheetForDirectory: path file: [downloadDirectoryButton title] types: nil modalForWindow: mainWindow modalDelegate: self didEndSelector: (SEL)@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo: sender];
}

- (IBAction)addToDatabasePath:(id)sender;
{
	NSOpenPanel * panel = [NSOpenPanel openPanel];
	NSString * path = nil;	// NSHomeDirectory();
	[panel setAllowsMultipleSelection: YES];
	[panel setCanChooseDirectories: YES];
	[panel setCanChooseFiles: NO];
	//	[NSApp stopModal];
	[panel beginSheetForDirectory: path file: nil types: nil modalForWindow: mainWindow modalDelegate: self didEndSelector: (SEL)@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo: sender];
}

- (IBAction)showIdentity:(id)sender
{
//	NSString * object;
//	NSArray * components;
//	NSSize size;

	//	[NSApp stopModal];
	NSAssert(identityView, @"identityView should not be nil");
//	object = [windowSizes objectForKey :@"identityView"];
//	NSAssert(object, @"windowSizes object (identityView) should not be nil");
//	components = [object componentsSeparatedByString: @" "];
//	NSAssert([components count] == 4, @"windowSizes components array (identityView) should contain four objects");
//	size.width = [[components objectAtIndex: 2] floatValue];
//	size.height = [[components objectAtIndex: 3] floatValue];
//	[mainWindow setContentSize: size];
	[mainView setContentView: identityView];
	[mainWindow center];
	[mainWindow makeKeyAndOrderFront: nil];
	[nickNameTextField setStringValue: [NSString stringWithCString: g_config->ReadString("nick","nickname")]];
	[realNameTextField setStringValue: [NSString stringWithCString: g_config->ReadString("userinfo","")]];
	[networkNameTextField setStringValue: [NSString stringWithCString: g_config->ReadString("networkname","")]];
	// Must select popup item by title instead
//	[connectionSpeedPopup selectItemAtIndex: g_config->ReadInt("keepupnet",0)];
//	[NSApp runModalForWindow: mainWindow];
}

- (BOOL)validateIdentity;
{
	bool valid = YES;
    if (![[nickNameTextField stringValue] length]) {
		NSRunAlertPanel(@"Missing Value", @"The 'Nickname' field cannot be empty.", @"OK", nil, nil);
		valid = NO;
	} else {
		strcpy(g_regnick, (char *)[[[nickNameTextField stringValue] substringWithRange: NSMakeRange(0, [[nickNameTextField stringValue] length] > 32 ? 32 : [[nickNameTextField stringValue] length])] cString]);
		g_config->WriteString("nick", g_regnick);
	}
    if (![[realNameTextField stringValue] length]) {
		NSRunAlertPanel(@"Missing Value", @"The 'Real name' field cannot be empty.", @"OK", nil, nil);
		valid = NO;
	} else {
		g_config->WriteString("userinfo", (char *)[[[realNameTextField stringValue] substringWithRange: NSMakeRange(0, [[realNameTextField stringValue] length] > 255 ? 255 : [[realNameTextField stringValue] length])] cString]);
	}
	NSString * s = [networkNameTextField stringValue];
	int l = [s length];
    if (l) {
		NSRunAlertPanel(@"Warning!", @"If 'Network name/ID' value is set, then you will be unable to see or connect to any hosts that do use the exact same ID.", @"OK", nil, nil);
	}
	g_config->WriteString("networkname", (char *)[[s substringWithRange: NSMakeRange(0, l > 255 ? 255 : l)] cString]);
	int conspeed_speeds[5]={32,64,384,1600,20000};
	g_config->WriteInt("conspeed", conspeed_speeds[[[connectionSpeedPopup selectedCell] tag]]);
	g_config->WriteInt("keepupnet", 8);
	g_config->Flush();
	return valid;
}

- (IBAction)showPassphrase:(id)sender
{
//	NSString * object;
//	NSArray * components;
//	NSSize size;

	if (![self validateIdentity]) {
		return;
	}

	//	[NSApp stopModal];
	NSAssert(passphraseView, @"passphraseView should not be nil");
//	object = [windowSizes objectForKey :@"passphraseView"];
//	NSAssert(object, @"windowSizes object (passphraseView) should not be nil");
//	components = [object componentsSeparatedByString: @" "];
//	NSAssert([components count] == 4, @"windowSizes components array (passphraseView) should contain four objects");
//	size.width = [[components objectAtIndex: 2] floatValue];
//	size.height = [[components objectAtIndex: 3] floatValue];
//	[mainWindow setContentSize: size];
	[mainView setContentView: passphraseView];
	[mainWindow center];
	[passphraseTextField selectText: nil];
	[mainWindow makeKeyAndOrderFront: nil];
	//	[NSApp runModalForWindow: mainWindow];
}

- (BOOL)validatePassphrase;
{
	bool valid = YES;
    if (![[passphraseTextField stringValue] length]) {
		NSRunAlertPanel(@"Missing Value", @"The 'Passphrase' field cannot be empty.", @"OK", nil, nil);
		valid = NO;
	}
    if (![[passphraseAgainTextField stringValue] length] || ![[passphraseTextField stringValue] isEqual: [passphraseAgainTextField stringValue]]) {
		NSRunAlertPanel(@"Passphrase Error", @"The passphase you entered did not match in both fields.", @"OK", nil, nil);
		valid = NO;
	}
    if ([[keySizePopup selectedCell] tag] > 2048) {
		NSRunAlertPanel(@"Caution!", @"Key sizes this large can take quite a while to generate.", @"OK", nil, nil);
	}
    if ([[keySizePopup selectedCell] tag] < 2048) {
		NSRunAlertPanel(@"Caution!", @"Key sizes this small may be cryptographically insecure.", @"OK", nil, nil);
	}
	return valid;
}

- (IBAction)showRandom:(id)sender
{
//	NSString * object;
//	NSArray * components;
//	NSSize size;

	if (![self validatePassphrase]) {
		return;
	}

	//	[NSApp stopModal];
	NSAssert(randomView, @"randomView should not be nil");

	if ([self needsRandomNumbers] == NO) {
		[self showKeygen: nil];
	}

	/*
	 running=1;
	 memset(&kg_privKey,0,sizeof(kg_privKey));
	 kg_thread=0;
	 kg_start_time=0;
	 memset(kg_movebuf,0,sizeof(kg_movebuf));
	 kg_movebuf_cnt=0;
	 kg_oldWndProc=0;
	 kg_going=1;
	 kg_keysize=2048;
	 memset(&kg_random,0,sizeof(kg_random));
	 sprintf(kg_privout,"%s.pr4",g_config_prefix);
	 NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: kg_privout] ofType: nil];
	 if (path) {
		 safe_strncpy(kg_privout,(char *)[path cString],sizeof(kg_privout));
	 }
#ifdef _DEBUG_
	 NSLog(@"%@ - showRandom : kg_privout = %s", [self class], kg_privout);
#endif
	 */
	[randomLabel setStringValue: @"Generating random numbers..."];
	[self setNoteString: @"Please move the mouse around randomly until the progress bar is filled."];
	[randomProgressBar setUsesThreadedAnimation: YES];
	[randomProgressBar setIndeterminate: NO];
	[randomProgressBar setMinValue: 0.0];
	[randomProgressBar setMaxValue: 100.0];
	[self setProgressBar: [NSNumber numberWithInt: 0]];
	[self setNeedsRandomNumbers: YES];
	[self setShouldQuit: NO];
	[randomCancelButton setEnabled: YES];
	[randomNextButton setEnabled: NO];

//	object = [windowSizes objectForKey :@"randomView"];
//	NSAssert(object, @"windowSizes object (randomView) should not be nil");
//	components = [object componentsSeparatedByString: @" "];
//	NSAssert([components count] == 4, @"windowSizes components array (randomView) should contain four objects");
//	size.width = [[components objectAtIndex: 2] floatValue];
//	size.height = [[components objectAtIndex: 3] floatValue];
//	[mainWindow setContentSize: size];
	[mainView setContentView: randomView];
	[mainWindow center];
//	[WasteApplication detachDrawingThread: (SEL)@selector(generateRandomAndKey) toTarget: self withObject: nil];
	[WasteApplication detachDrawingThread: (SEL)@selector(generateRandomAndKey) toTarget: self withObject: nil];
	[mainWindow makeKeyAndOrderFront: nil];
//	[WasteApplication runModalForWindow: mainWindow];
}

- (IBAction)showKeygen:(id)sender
{
//	NSString * object;
//	NSArray * components;
//	NSSize size;

	//	[NSApp stopModal];
	NSAssert(keygenView, @"keygenView should not be nil");
//	object = [windowSizes objectForKey: @"keygenView"];
//	components = [object componentsSeparatedByString: @" "];
//	size.width = [[components objectAtIndex: 2] floatValue];
//	size.height = [[components objectAtIndex: 3] floatValue];
//	[mainWindow setContentSize: size];
	[mainView setContentView: keygenView];
	[mainWindow center];
	[mainWindow makeKeyAndOrderFront: nil];
	//	[NSApp runModalForWindow: mainWindow];
}

- (IBAction)showProfile:(id)sender
{
//	NSString * object;
//	NSArray * components;
//	NSSize size;

	//	[NSApp stopModal];
//	object = [windowSizes objectForKey :@"profileView"];
//	components = [object componentsSeparatedByString: @" "];
//	size.width = [[components objectAtIndex: 2] floatValue];
//	size.height = [[components objectAtIndex: 3] floatValue];
//	[mainWindow setContentSize: size];
	[mainView setContentView: profileView];
	[mainWindow center];
	[downloadDirectoryButton setTitle: [NSString stringWithCString: g_config->ReadString("downloadpath","")]];
	[databasePathTextField setStringValue: [NSString stringWithCString: g_config->ReadString("databasepath","")]];
	[mainWindow makeKeyAndOrderFront: nil];
	//	[NSApp runModalForWindow: mainWindow];
}

- (BOOL)validateProfile;
{
	bool valid = YES;
    if (![[databasePathTextField stringValue] length]) {
		NSRunAlertPanel(@"Missing Value", @"The 'Database Path' field cannot be empty.\nNote: You may leave the directory itself devoid of files.", @"OK", nil, nil);
		valid = NO;
	}
	NSString * s = [downloadDirectoryButton title];
	int l = [s length];
    if (!l) {
		NSRunAlertPanel(@"Missing Value", @"You must select a 'Download Directory' before continuing.", @"OK", nil, nil);
		valid = NO;
	}
	g_config->WriteString("downloadpath", (char *)[[s substringWithRange: NSMakeRange(0, l > 4095 ? 4095 : l)] cString]);
	s = [databasePathTextField stringValue];
	l  = [s length];
	g_config->WriteString("databasepath", (char *)[[s substringWithRange: NSMakeRange(0, l > 4095 ? 4095 : l)] cString]);
	g_config->Flush();
	return valid;
}

- (void)generateRandomAndKey0;
{
	NSAssert(g_config_prefix, @"g_config_prefix should not ne nil");
	NSAssert(kg_privout, @"kg_privout should not ne nil");
	sprintf(kg_privout,"%s.pr4",g_config_prefix);
	NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: kg_privout] ofType: nil];
	if (path) {
		safe_strncpy(kg_privout,(char *)[path cString],sizeof(kg_privout));
	}

#ifdef _DEBUG_
	NSLog(@"%@ - showRandom : kg_privout = %s", [self class], kg_privout);
#endif

	kg_keysize = [[keySizePopup selectedCell] tag];

	const char * passwordPointer = [[passphraseTextField stringValue] cString];
#ifdef _DEBUG_
	NSLog(@"%@ - generateRandomAndKey : passwordPointer = %s", [self class], (unsigned char *)passwordPointer);
#endif
	SHAify c;
	c.add((unsigned char *)passwordPointer, strlen(passwordPointer));
	c.final((unsigned char *)kg_passbf);
	[passphraseAgainTextField setStringValue: @""];
	[passphraseTextField setStringValue: @""];
#ifdef _DEBUG_
	NSData * data = [NSData dataWithBytes: kg_passbf length: sizeof(kg_passbf)];
	NSLog(@"%@ - generateRandomAndKey : kg_passbf = %@", [self class], data);
#endif

	[randomLabel setStringValue:  @"Generating key pair..."];
	[self setNoteString: [NSString stringWithFormat: @"Generating the RSA key pairs may take as long as %d minute%s.", kg_keysize<=1024?1:kg_keysize<=2048?3:kg_keysize<=3072?10:20, kg_keysize==1024?"":"s"]];
	[randomProgressBar setIndeterminate: NO];
	[randomProgressBar startAnimation: nil];
	[randomCancelButton setEnabled: NO];
	[randomNextButton setEnabled: NO];
	NSCalendarDate * startTime = [NSCalendarDate calendarDate];
	[mainWindow makeKeyAndOrderFront: nil];

	keygen([self keygenLock]);

	NSCalendarDate * stopTime = [NSCalendarDate calendarDate];
	NSTimeInterval t = [stopTime timeIntervalSinceDate: startTime];
	double minutes, seconds;
	seconds = fmod(t, 60);
	minutes = t / 60;
	[randomLabel setStringValue:  @""];
	[self setNoteString: [NSString stringWithFormat: @"RSA Key generation complete.\nElapsed time : %d minute%s : %f seconds", (int)minutes, (int)minutes==1?"":"s", seconds]];
	[randomProgressBar stopAnimation: nil];
	[randomProgressBar setDoubleValue: 100.0];
	[randomProgressBar setIndeterminate: NO];
	kg_going = 0;

	running=0;
	if ([self shouldQuit]) {
		[self setNeedsRandomNumbers: YES];
//		[NSApp stopModal];
		return;
	}
//	[NSApp stopModal];
	[WasteApplication detachDrawingThread: (SEL)@selector(finishProgressBar) toTarget: self withObject: nil];
}

- (void)generateRandomAndKey;
{
	 NSPoint point, currentPoint;
	 unsigned int bytesNeeded;
	 unsigned char buf[16];
	 const char *passwordPointer;
	 FILE *fp;

	 NSAssert(g_config_prefix, @"g_config_prefix should not ne nil");
	 NSAssert(kg_privout, @"kg_privout should not ne nil");
	 sprintf(kg_privout,"%s.pr4",g_config_prefix);
	 NSString * path = [[NSBundle mainBundle] pathForResource: [NSString stringWithCString: kg_privout] ofType: nil];
	 if (path) {
		 safe_strncpy(kg_privout,(char *)[path cString],sizeof(kg_privout));
	 }

	 kg_keysize = [[keySizePopup selectedCell] tag];

	 passwordPointer = [[passphraseTextField stringValue] cString];
	 SHAify c;
	 c.add((unsigned char *)passwordPointer, strlen(passwordPointer));
	 c.final((unsigned char *)kg_passbf);
	 [passphraseAgainTextField setStringValue: @""];
	 [passphraseTextField setStringValue: @""];
#ifdef _DEBUG_
//	 NSData * data = [NSData dataWithBytes: kg_passbf length: sizeof(kg_passbf)];
//	 NSLog(@"%@ - generateRandomAndKey : kg_passbf = %@", [self class], data);
#endif

	 R_RandomInit(&kg_random);
	 fp=fopen("/dev/random","rb");
	 if (fp)	{
		 fread(kg_random.state,sizeof(kg_random.state),1,fp);
		 fclose(fp);
		 kg_random.outputAvailable=0;
		 kg_random.bytesNeeded=0;
		 kg_going = 2;
	 } else {
		 kg_random.bytesNeeded=sizeof(kg_random.state);
		 R_GenerateBytes(buf,sizeof(buf),&g_random);
		 R_RandomUpdate(&kg_random,buf,sizeof(buf));
		 memset(buf,0,sizeof(buf));
	 }
	 while (kg_going && ![self shouldQuit])
	 {
		 if (kg_going == 1) {
			 point = [mainWindow mouseLocationOutsideOfEventStream];
			 currentPoint = [mainWindow mouseLocationOutsideOfEventStream];
			 while (point.x == currentPoint.x && point.y == currentPoint.y) {
				 currentPoint = [mainWindow mouseLocationOutsideOfEventStream];
			 }
//			int j, n = 1000;
//			for (j = 0; j < n; j++) {
				point = [mainWindow mouseLocationOutsideOfEventStream];
			 [NSThread sleepUntilDate: [NSDate dateWithTimeIntervalSinceNow: 0.1]];
//			}
//			NSLog(@"%@ - generateRandomAndKey : kg_movebuf_cnt = %d", [self class], kg_movebuf_cnt);
//			NSLog(@"%@ - generateRandomAndKey : mouse currentPoint = %f, %f", [self class], currentPoint.x, currentPoint.y);
			 kg_movebuf[kg_movebuf_cnt%7]+=(unsigned int)(fabs(currentPoint.x) * fabs(currentPoint.y));
			 kg_movebuf[(kg_movebuf_cnt+1)%7]+=(unsigned int)[[NSDate date] timeIntervalSince1970];
			 point = [mainWindow mouseLocationOutsideOfEventStream];
//			NSLog(@"%@ - generateRandomAndKey : mouse point = %f, %f", [self class], point.x, point.y);
			 kg_movebuf[(kg_movebuf_cnt+2)%7]+=(unsigned int)(fabs(point.x) * fabs(point.y));
//			NSLog(@"%@ - generateRandomAndKey : mouse point = %f, %f", [self class], point.x, point.y);
			 [self setProgressBar: [NSNumber numberWithInt: (int)(100-(bytesNeeded/2.12))]];
			 if (++kg_movebuf_cnt >= 29)
			 {
//				NSLog(@"%@ - generateRandomAndKey : kg_movebuf_cnt = %d", [self class], kg_movebuf_cnt);
				 kg_movebuf_cnt=0;
				 R_RandomUpdate(&kg_random,(unsigned char *)kg_movebuf,sizeof(kg_movebuf));

				 R_GetRandomBytesNeeded(&bytesNeeded, &kg_random);
				 [self setProgressBar: [NSNumber numberWithInt: (int)(100-(bytesNeeded/2.12))]];
//				NSLog(@"%@ - generateRandomAndKey : bytesNeeded = %d", [self class], bytesNeeded);
				 if (!bytesNeeded)
				 {
//					[NSApp stopModal];
					 [randomLabel setStringValue:  @""];
					 kg_going=2;
//					NSLog(@"%@ - generateRandomAndKey : Generating key pair... please wait", [self class]);
					 [self setProgressBar: [NSNumber numberWithInt: 100]];
				 }
			 }
		 }
		 if (kg_going == 2) {
			 [randomLabel setStringValue:  @"Generating key pair..."];
			 [self setNoteString: [NSString stringWithFormat: @"Generating the RSA key pairs may take as long as %d minute%s.", kg_keysize<=1024?1:kg_keysize<=2048?3:kg_keysize<=3072?10:20, kg_keysize==1024?"":"s"]];
			 [randomProgressBar setIndeterminate: NO];
			 [randomProgressBar startAnimation: nil];
			 [randomCancelButton setEnabled: NO];
			 [randomNextButton setEnabled: NO];
			 NSCalendarDate * startTime = [NSCalendarDate calendarDate];
			 [mainWindow makeKeyAndOrderFront: nil];
			 if (keygen([self keygenLock])) {
				 NSRunAlertPanel(@"Key Generation Error", @"Error create keys or write private key file", @"Cancel", nil, nil);
				 [NSApp terminate: self];
			 }
			 NSCalendarDate * stopTime = [NSCalendarDate calendarDate];
			 NSTimeInterval t = [stopTime timeIntervalSinceDate: startTime];
			 double minutes, seconds;
			 seconds = fmod(t, 60);
			 minutes = t / 60;
			 [randomLabel setStringValue:  @""];
			 [self setNoteString: [NSString stringWithFormat: @"RSA Key generation complete.\nElapsed time : %d minute%s : %f seconds", (int)minutes, (int)minutes==1?"":"s", seconds]];
			 [randomProgressBar stopAnimation: nil];
			 [randomProgressBar setDoubleValue: 100.0];
			 [randomProgressBar setIndeterminate: NO];
			 kg_going = 0;
		 }
	 }

	 running=0;
	 if ([self shouldQuit]) {
		 [self setNeedsRandomNumbers: YES];
//		[NSApp stopModal];
		 return;
	 }

//	[NSApp stopModal];
	 [WasteApplication detachDrawingThread: (SEL)@selector(finishProgressBar) toTarget: self withObject: nil];
}

- (void)finishProgressBar;
{
	[self setProgressBar: [NSNumber numberWithInt: 100]];
	[randomCancelButton setEnabled: NO];
	[randomNextButton setEnabled: YES];
	//	[self setNeedsRandomNumbers: NO];
 //	[NSApp performSelector: @selector(runModalForWindow:) withObject: mainWindow afterDelay: 1.0];
	[mainWindow display];
	[mainWindow makeKeyAndOrderFront: nil];
}

- (void)setProgressBar: (NSNumber *) number;
{
	[randomProgressBar setDoubleValue: [number doubleValue]];
}

- (void)setNoteString: (NSString *) string;
{
	[randomNote setStringValue: string];
}

- (void)openPanelDidEnd: (NSOpenPanel *)sheet returnCode: (int)returnCode contextInfo: (void  *)contextInfo;
{
	if (downloadDirectoryButton == contextInfo) {
		if (returnCode == NSAlertDefaultReturn) {
			if ([[sheet filenames] count]) {
				[downloadDirectoryButton setTitle: [[sheet filenames] objectAtIndex: 0]];
			}
		}
	}
	if (databasePathAddButton == contextInfo) {
		if (returnCode == NSAlertDefaultReturn) {
			if ([[sheet filenames] count]) {
				NSString * currentValue = [databasePathTextField stringValue];
				if ([currentValue length]) {
					[databasePathTextField setStringValue: [NSString stringWithFormat: @"%@;%@", currentValue, [[sheet filenames] objectAtIndex: 0]]];
				} else {
					[databasePathTextField setStringValue: [[sheet filenames] objectAtIndex: 0]];
				}
			}
		}
	}
	if (keygenImportPublicKeyButton == contextInfo) {
		if (returnCode == NSAlertDefaultReturn) {
			if ([[sheet filenames] count]) {
				[self importPublicKeysFromFile: [[sheet filenames] objectAtIndex: 0]];
			}
		}
	}
	[mainWindow display];
	[mainWindow makeKeyAndOrderFront: nil];
}


- (void)dealloc;
{
#ifdef _DEBUG_
	NSLog(@"%@ - dealloc", [self class]);
#endif
	[identityView release];
	[passphraseView release];
	[randomView release];
	[keygenView release];
	[profileView release];
//	[windowSizes release];
	[keygenLock release];
	[super dealloc];
}

@end
