#import "WasteApplication.h"

#import "ClientController.h"
#import "ChatController.h"

extern void debug_printf(char *, ...);
extern int g_exit;
extern int g_has_started;
extern int g_scanworking;
extern char g_scan_status_buf[128];

extern int waste_main(id theController);
extern void stop_waste(void);
//static R_RANDOM_STRUCT joke_random;

#define APPAutoStartServerOnLaunch @"autoStartServerOnLaunch"

@implementation WasteApplication

+ (void)initialize;
{
	NSUserDefaults * defaults = [NSUserDefaults standardUserDefaults];
	NSDictionary * appDefaults = [NSDictionary dictionaryWithObject: @"YES" forKey: APPAutoStartServerOnLaunch];
	[defaults registerDefaults: appDefaults];
}

+ (WasteApplication *)sharedApplication;
{
	return (WasteApplication *)[super sharedApplication];
}

- (id)init;
{
	self = [super init];
	[self setDelegate: self];
	return self;
}

- (ClientController *) clientController;
{
	return clientController;
}

- (SetupController *) setupController;
{
	if (!setupController) {
		[NSBundle loadNibNamed: @"SetupWizard.nib" owner: self];
		NSAssert(setupController, @"setupController should not be nil");
	}
	return setupController;
}

- (ChatController *) chatController;
{
	if (!chatController) {
		chatController = [[ChatController alloc] init];
//		[chatController setAppController: self];
	}
	return chatController;
}

/////////////////////////////////////////////////////////////////////////////////////
// Method used to get passphrase for existing private key
/////////////////////////////////////////////////////////////////////////////////////

- (const char *)getPassphrase;
{
	debug_printf("getPassphrase\n");
	[NSBundle loadNibNamed: @"PassphrasePanel.nib" owner: self];
	NSAssert(passphraseController, @"passphraseController should not be nil");
	return ([passphraseController _getPassphrase]);
}

//////////////////////////////////////////////////////////////////////////////
// Internal action methods
//////////////////////////////////////////////////////////////////////////////

- (void) startWASTE;
{
	debug_printf("startWASTE\n");
	[WasteApplication detachDrawingThread: (SEL)@selector(spawnServer) toTarget: self withObject: nil];
//	[NSThread detachNewThreadSelector: (SEL)@selector(spawnServer) toTarget: self withObject: nil];
}

- (void) stopWASTE;
{
	stop_waste();
//	[NSThread sleepUntilDate: [[NSCalendarDate calendarDate] dateByAddingYears: 0 months: 0 days: 0 hours: 0 minutes: 0 seconds: 1]];
}

- (void) spawnServer;
{
	NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	int retval;
	debug_printf("%@ - spawnServer\n");
	g_has_started = 1;
	is_setting_up = 0;
	g_exit = 0;
	[clientController startSpinner];
	retval = waste_main(self);
	g_has_started = 0;
	debug_printf("spawnServer : waste_main() exited with value %d\n", retval);
	NSAssert(clientController, @"clientController should not be nil");
	if (retval == 1)	// Setup needed
	{
		[WasteApplication detachDrawingThread: (SEL)@selector(performSetup) toTarget: self withObject: nil];
	}
	[pool release];
	return;
}

- (void) gotNick: (NSString *) nick;
{
	NSAssert(clientController, @"clientController should not be nil");
	[clientController gotNick: nick];
}

- (void) gotChannel: (NSString *) channel;
{
	NSAssert(clientController, @"clientController should not be nil");
	[clientController gotChannel: channel];
}

- (void) updateStatus;
{
	if (g_has_started && !g_exit) {
		[clientController updateServerStatus];
	}
}

/////////////////////////////////////////////////////////////////////////////////////
// Setup method used to invoke first-time-use private key generation
/////////////////////////////////////////////////////////////////////////////////////

- (void)performSetup;
{
	debug_printf("performSetup\n");
	[NSBundle loadNibNamed: @"SetupWizard.nib" owner: self];
	NSAssert(setupController, @"setupController should not be nil");
	[setupController showIdentity: nil];
	[self setIsSettingUp: YES];
}

- (void)setIsSettingUp: (BOOL)aFlag;
{
	is_setting_up = aFlag;
}

/////////////////////////////////////////////////////////////////////////////////////
// IBAction methods
/////////////////////////////////////////////////////////////////////////////////////

- (IBAction) showConsole:(id)sender;
{
	[[self clientController] showConsole];
}

- (IBAction) showJoinChat:(id)sender;
{
	[[self clientController] joinChat];
}

- (IBAction) showUsers:(id)sender;
{
	[mainWindow makeKeyAndOrderFront: nil];
}

- (IBAction) showStats:(id)sender;
{
	if (!statsController) {
		[NSBundle loadNibNamed: @"NetStats.nib" owner: self];
	}
	NSAssert(statsController, @"statsController should not be nil");
	[statsController showStats: self];
}

- (IBAction)setup:(id)sender;
{
	[self performSetup];
}

- (IBAction)start:(id)sender;
{
// Stopping the server does work, but restarting it again crashes the app, probably due to a freed resource.
// However since this application is (as the Wintel client/server) a combination client/server, I reasoned that
// it did not make sense to allow the user to stop the server discretely, vs. simply shutting down the app.
//	[sender setAction: (SEL)@selector(stop:)];
//	[sender setTitle: @"Stop Server"];
	[sender setEnabled: NO];
	if (!g_has_started) {
		strcpy(g_scan_status_buf, "Server Starting...");
		[clientController updateStatus];
	} else {
		strcpy(g_scan_status_buf, "Server Restarting...");
		[clientController updateStatus];
		g_exit = 1;
		while (g_has_started) {	// Wait for the server to shut down.
			[NSThread sleepUntilDate: [[NSCalendarDate calendarDate] dateByAddingYears: 0 months: 0 days: 0 hours: 0 minutes: 0 seconds: 1]];
		}
	}
	[self startWASTE];
}

// See note in -start: method above
- (IBAction)stop:(id)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - stop: %@", [self class], sender);
#endif
	[sender setAction: (SEL)@selector(login:)];
	[sender setTitle: @"Start Server"];
//	if (g_has_started) {
		[self stopWASTE];
//	}
}

 - (IBAction)setPrefs:(id)sender;
{
//	NSUserDefaults * defaults = [NSUserDefaults standardUserDefaults];
//	bool value = [prefsAutoStartButton state];
//	[prefsPanel close];
#ifdef _DEBUG_
	NSLog(@"%@ - setPrefs: %@", [self class], sender);
#endif
	[NSBundle loadNibNamed: @"Preferences.nib" owner: self];
	NSAssert(prefsController, @"prefsController should not be nil");
	[prefsController showPrefs: self];
//	[defaults setBool: value forKey: APPAutoStartServerOnLaunch];
}

- (IBAction)showTransfers: (id)sender;
{
	[clientController showTransfers];
}

- (IBAction)showBrowser: (id)sender;
{
	[clientController showBrowser];
}

- (IBAction)showNetworkStatus: (id)sender;
{
	[clientController showNetworkStatus];
}

- (void)getPrefs;
{
//	NSUserDefaults * defaults = [NSUserDefaults standardUserDefaults];
//	bool value = [defaults boolForKey: APPAutoStartServerOnLaunch];
//	[prefsAutoStartButton setState: value];
#ifdef _DEBUG_
	NSLog(@"%@ - getPrefs", [self class]);
#endif
}

- (NSString *)publicKey;
{
	return publicKey;
}

- (void)setPublicKey: (NSString *)theKey;
{
	[publicKey release];
	publicKey = [theKey retain];
}

- (void)copyPublicKeyToPasteboard: (NSString *)theKey;
{
	NSPasteboard * pb = [NSPasteboard generalPasteboard];
	[self setPublicKey: theKey];
	[pb declareTypes: [NSArray arrayWithObject: NSStringPboardType] owner: self];
}

- (void)pasteboard:(NSPasteboard *)sender provideDataForType:(NSString *)type;
{
	[sender setString: [self publicKey] forType: type];
}

- (void)dealloc;
{
	[activityTimer invalidate];
	[activityTimer release];
	[publicKey release];
	[super dealloc];
}

//////////////////////////////////////////////////////////////////////////////
// NSApplication delegate methods
//////////////////////////////////////////////////////////////////////////////

- (NSApplicationTerminateReply)applicationShouldTerminate: (NSApplication *)sender;
{
#ifdef _DEBUG_
	NSLog(@"%@ - applicationShouldTerminate: %@", [self class], sender);
#endif

	// If WASTE has not yet started (during first-time private keygen for instance)
	// don't bother waiting, just die right away.
	if (!g_has_started) {
		return NSTerminateNow;
	}

	stop_waste();

	[NSThread sleepUntilDate: [[NSCalendarDate calendarDate] dateByAddingYears: 0 months: 0 days: 0 hours: 0 minutes: 0 seconds: 1]];

	// Tell the application not to terminate just yet
	// The actual termination method is invoked by the WASTE C++ code after flushing the DB
	return NSTerminateLater;
}

- (void)applicationDidFinishLaunching: (NSNotification *)notification;
{
#ifdef _DEBUG_
	NSLog(@"%@ - applicationDidFinishLaunching: %@", [self class], notification);
#endif
//	[self getPrefs];
	[NSBundle loadNibNamed: @"Client.nib" owner: self];
	NSAssert(clientController, @"clientController should not be nil");
	activityTimer = [[NSTimer scheduledTimerWithTimeInterval: 0.1 target: self selector: (SEL)@selector(updateStatus) userInfo: nil repeats: YES] retain];
//	[activityTimer fire];
	strcpy(g_scan_status_buf, "Server Not Started");
	[clientController updateStatus];
	[self start: nil];
//	[[SharedAppController sharedController] _setAppController: self];
}

- (void)applicationDidBecomeActive:(NSNotification *)notification;
{
	[mainWindow makeKeyAndOrderFront: self];
}

- (void)applicationDidUnhide:(NSNotification *)notification;
{
	[mainWindow makeKeyAndOrderFront: self];
}

//////////////////////////////////////////////////////////////////////////////
// Protocol methods for validating menu items
//////////////////////////////////////////////////////////////////////////////

- (BOOL)validateUserInterfaceItem:(id <NSValidatedUserInterfaceItem>)anItem;
{
	return [self shouldShowMenuItem: anItem];
}

- (BOOL)shouldShowMenuItem: (id)anItem;
{
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Network Status..."] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Browser..."] && !is_setting_up) {
			return NO;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Show Console..."] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Transfers..."] && !is_setting_up) {
			return NO;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Waste Help"] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Current Users..."] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Create/Join Chat..."] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Network Statistics..."] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Setup..."] && !is_setting_up) {
			if (!g_key.bits) {
				return YES;
			}
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Quit Waste"] && !is_setting_up) {
			return YES;
		}
	}
	if ([anItem respondsToSelector: (SEL)@selector(title)]) {
		if ([[anItem title] isEqualToString: @"Preferences..."] && !is_setting_up) {
			return YES;
		}
	}
	if (is_setting_up) {
		return NO;
	}
	return YES;
}

//////////////////////////////////////////////////////////////////////////////
// NSWindow delegate methods
//////////////////////////////////////////////////////////////////////////////

//- (BOOL)windowShouldClose:(id)sender;
//{
//	[NSApp terminate: sender];
//	return YES;
//}

@end
