/*
 WASTE - platform.h (Windows / Posixish platform includes and defines)
 Copyright (C) 2003 Nullsoft, Inc.

 WASTE is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 WASTE  is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with WASTE; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _PLATFORM_H_
#define _PLATFORM_H_

/* UINT2 defines a two byte word */
typedef unsigned short int UINT2;

/* UINT4 defines a four byte word */
typedef unsigned int UINT4;

#ifdef _WIN32

#include <windows.h>
#include <windowsx.h>
#include <richedit.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <process.h>
#include <time.h>
#include <commctrl.h>
#include <shlobj.h>
#include <ctype.h>

#define DIRCHAR '\\'

//define if you want debug output in the system event log
#undef SYSTEM_EVENT_LOG

#define sprintf wsprintf
#define safe_strncpy lstrcpyn
#define socklen_t int

#else

//#define THREAD_SAFE
#define _REENTRANT
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#define DIRCHAR '/'
#define CharNext(x) (x+1)
#define CharPrev(s,x) ((s)<(x)?(x)-1:(s))
#define DeleteFile(x) unlink(x)
#define CreateDirectory(x,y) mkdir(x,0755)
#define RemoveDirectory(x) rmdir(x)
#define Sleep(x) usleep(( x )*1000)
#define stricmp(x,y) strcasecmp(x,y)
#define strnicmp(x,y,z) strncasecmp(x,y,z)
static inline char *safe_strncpy(char *out, char *in, int maxl) { return strncpy(out,in,maxl); out[maxl-1]=0; }

static inline unsigned int GetTickCount()
{
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return (tv.tv_sec * 1000) + (tv.tv_usec/1000);
}

#define MAX_PATH 1024

#endif

#ifndef min
#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((y)<(x)?(x):(y))
#endif

#if defined(__FreeBSD__) || defined(__APPLE__)
#include <machine/endian.h>
#endif
#if defined(__linux__)
#include <endian.h>
#endif
#ifndef BIG_ENDIAN
#define BIG_ENDIAN 4321
#endif
#ifndef LITTLE_ENDIAN
#define LITTLE_ENDIAN 1234
#endif
#if defined(sparc) || defined(__sparc) || defined(__sparc__)
#define BYTE_ORDER BIG_ENDIAN
#endif
#ifdef _WIN32
#define BYTE_ORDER LITTLE_ENDIAN
#endif

#if !defined(_WIN32) && !defined(__APPLE__)
#define USE_POLL
#endif

#ifndef INADDR_NONE
#define INADDR_NONE ((in_addr_t) -1)
#endif

#if defined(__linux__) || defined(__FreeBSD__) || defined(_DIRENT_HAVE_D_TYPE)
#define HAVE_STRUCT_DIRENT_D_TYPE
#else
#define DT_DIR 4
#endif

#ifdef _WIN32
#define INT64 __int64
#endif
#ifdef __GNUC__
#define INT64 long long
#endif

#if defined(sparc) || defined(__sparc) || defined(__sparc__) || defined(__APPLE__)
/* Blowfish encryption/decryption dereferences arrays of char as UINT4s.
* Some platforms (eg. Solaris Ultra-2) require strict alignment.
*    pgar, 16 June 2003
*/
#define USE_STRICT_ALIGNMENT
#endif

#define PGAR_BPS

#endif // _PLATFORM_H_
