/*
 WASTE - sockets.h (some socket portability defines)
 Copyright (C) 2003 Nullsoft, Inc.

 WASTE is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 WASTE  is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with WASTE; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 -- this really should be in platform.h, but oh well :)

 */

#ifndef _SOCKETS_H_
#define _SOCKETS_H_

#ifdef _WIN32
#define in_addr_t unsigned long
#define ERRNO (WSAGetLastError())
#define SET_SOCK_BLOCK(s,block) { unsigned long __i=block?0:1; ioctlsocket(s,FIONBIO,&__i); }
#define EWOULDBLOCK WSAEWOULDBLOCK
#define EINPROGRESS WSAEWOULDBLOCK

#else
#define ERRNO errno
#define closesocket(s) close(s)
#define SET_SOCK_BLOCK(s,block) { int __flags; if ((__flags = fcntl(s, F_GETFL, 0)) != -1) {  if (!block) __flags |= O_NONBLOCK; else __flags &= ~O_NONBLOCK; fcntl(s, F_SETFL, __flags); } }

#endif

#if defined(__APPLE__)
#define in_addr_t u_int32_t
#ifndef socklen_t
#define socklen_t int
#endif
#endif

#define HAVE_INET_ATON
#if defined(_WIN32) || defined(sun)
#undef HAVE_INET_ATON
#define HAVE_INET_ADDR
#endif

#if !defined(HAVE_INET_ATON)
inline int inet_aton(const char *cp, struct in_addr *inp)
{
#if defined(HAVE_INET_ADDR)
	inp->s_addr = inet_addr(cp);
	return inp->s_addr!=INADDR_NONE;
#else
	int b1,b2,b3,b4;
	sscanf(cp,"%d.%d.%d.%d",&b1,&b2,&b3,&b4);
	if ((b1<0)||(b1>255)||(b2<0)||(b2>255)||(b3<0)||(b3>255)||(b4<0)||(b4>255))
		return 0;
	inp->S_un.S_un_b.s_b1 = b1;
	inp->S_un.S_un_b.s_b2 = b2;
	inp->S_un.S_un_b.s_b3 = b3;
	inp->S_un.S_un_b.s_b4 = b4;
	return 1;
#endif
}
#endif

#ifndef SHUT_RDWR
#define SHUT_RDWR 2
#endif

#endif //_SOCKETS_H_
